#ifndef BNX_H
#define BNX_H

#include <sstream>
#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <map>
#include <chrono>
#include "bnxrow.hpp"
#include "../Configuration/staticconf.hpp"

using namespace std;

// class representing one BNX file
class Bnx
{
  public:
    // constructor
    Bnx(string filePath);
    
    // destructor
    ~Bnx();

    // fills this BNX object with data from file set in BnxFilePath
    // if 0 or negative number is provided as parameter then all rows from the file are read
    bool Fill(long rowsToRead, bool printStatistics);

    // path to BNX file
    string BnxFilePath;
    
    // method which provides mapping from string to BNX row type enum
    BnxRowType GetBnxRowType(string row);

    // map containing raw categorized data by theirs BnxRowType
    map<BnxRowType, vector<BnxRow>*>* _rows;
};

Bnx::Bnx(string filePath)
{
  _rows = 0;
  BnxFilePath = filePath;
}

Bnx::~Bnx()
{
  for (auto const& ri : (*_rows))
  {
    ri.second->clear();
    delete ri.second;
  }

  delete _rows;
}

bool Bnx::Fill(long rowsToRead, bool printStatistics)
{
  auto start = chrono::steady_clock::now();

  if (BnxFilePath.length() == 0)
  {
    throw "File path must be specified!";
  }

  if (printStatistics)
  {
    cout << "Start parsing file: " << BnxFilePath << "\n";
    cout << "\nProgress: ";
  }

  if (_rows == 0)
  {
    _rows = new map<BnxRowType, vector<BnxRow>*>();
  }

  ifstream infile(BnxFilePath);
  string fileRow;
  long progress = 0;

  // do the read itslef
  while (getline(infile, fileRow))
  {
    if (printStatistics && progress % 100000 == 0)
    {
      cout << "*";
    }

    // get the type of row and push it into the proper vector
    auto rowType = GetBnxRowType(fileRow);
    auto rowData = new vector<string>();
    
    // do not separate the header rows, just store it in first vector field
    if (rowType == BnxRowType::BnxFileHeader || rowType == BnxRowType::Undefined)
    {
      rowData->push_back(fileRow);
    }
    else
    {
      // remove line type string from the beginning of the string (plus 1 for space after the string)
      auto strToRemove = GetBnxRowTypeNames()[rowType];
      fileRow.erase(0, strToRemove.length() + 1);
      
      // separate the row string by tabs and store it in the vector
      int s = 0;
      auto fileRowLen = fileRow.length();
      for (int i = 0; i <= fileRowLen; i++)
      {
        if (fileRow[i] == '\t' || i == fileRowLen)
        {
          if (s != i)
          {
            auto substr = fileRow.substr(s, i - s);
            rowData->push_back(substr);
          }
          
          s = i + 1;
        }
      }
    }

    if ((*_rows)[rowType] == 0)
    {
      (*_rows)[rowType] = new vector<BnxRow>();
    }

    (*_rows)[rowType]->push_back(rowData);

    if (rowsToRead > 0 && progress == rowsToRead)
    {
      break;
    }

    progress++;
  }

  infile.close();
  
  // print out some statistics
  if (printStatistics)
  {
    cout << "\n\nFile parsing finished.\n";
    cout << "\nBNX rows statistics:\n";
    for (auto const& r : (*_rows))
    {
      cout << GetBnxRowTypeNames()[r.first] << " => " << r.second->size() << "\n";
    }

    auto end = chrono::steady_clock::now();
    chrono::duration<double> diff = end - start;
    cout << "\nTime to process BNX file: " << diff.count() << "s\n\n";
  }

  return true;
}

BnxRowType Bnx::GetBnxRowType(string row)
{
  if (row.length() == 0)
  {
    return BnxRowType::Undefined;
  }

  // most common rows in BNX files starts with 0, 1 or Q
  // so handle it separately by direct indexing
  switch (row[0])
  {
    case '0':
      return BnxRowType::Data0;
    case '1':
      return BnxRowType::Data1;
    case 'Q':
      if (row[1] == '1' && row[2] == '1')
        return BnxRowType::Quality11;
      else if (row[1] == '1' && row[2] == '2')
        return BnxRowType::Quality12;
      else if (row[1] == '2' && row[2] == '1')
        return BnxRowType::Quality21;
      else if (row[1] == '2' && row[2] == '2')
        return BnxRowType::Quality22;
    default:
      break;
  }

  // do the rest of the rows
  for (auto const& m : GetStrToBnxRowTypeMap())
  {
    if (row.find(m.first) == 0)
    {
      return m.second;
    }
  }

  return BnxRowType::Undefined;
}

#endif