#ifndef RUNSDATA_H
#define RUNSDATA_H

#include <vector>
#include <map>
#include <climits>
#include "run.hpp"
#include "bnx.hpp"
#include "bnxrow.hpp"
#include "../Configuration/staticconf.hpp"

template<typename T>
struct RelativeLabel
{
  int Count;
  vector<T> ApproxLabels;
};

template<typename T>
class RunsData
{
  public:
    // constructor
    RunsData();

    // runs data map, key is RunID
    std::map<int, Run<T>*> _data;

    // distinct list of label distances
    // key is label relative distance, value is relative label struct
    std::map<int, RelativeLabel<T>> _labelDistances;

    T _maxLabelDistance;
    
    // fills RunsData with data from provided BNX file
    // if minMoleculeLabelsCnt is provided than all molecules with less than this amount will be left out (length filtering)
    void FillDataFromBnx(Bnx* bnxFile, bool generateReverseMolecules, int minMoleculeLabelsCnt, bool printStatistics);

    // calculates and prints statistics about runs
    void PrintStatistics();
};

template<typename T>
RunsData<T>::RunsData()
{
  std::map<int, Run<T>*> _data;
  std::map<int, RelativeLabel<T>> _labelDistances;
  _maxLabelDistance = 0;
}

template<typename T>
void RunsData<T>::FillDataFromBnx(Bnx* bnxFile, bool generateReverseMolecules, int minMoleculeLabelsCnt, bool printStatistics)
{
  auto start = chrono::steady_clock::now();

  if (printStatistics)
  {
    cout << "Start of BNX data parsing...\n";
  }

  // create run objects first by iterating through data rows about runs
  for (auto const& ri : (*(*bnxFile->_rows)[BnxRowType::RunData]))
  {
    auto runId = stoi(ri.RowData->back());
    _data[runId] = new Run<T>(runId);
  }

  // just check if we have same number of 0 and 1 data rows parsed as they must always be in pairs
  if ((*bnxFile->_rows)[BnxRowType::Data0]->size() != (*bnxFile->_rows)[BnxRowType::Data1]->size())
  {
    throw "There is different number of header and data rows! Can't proceed.";
  }

  // iterate through the molecules and create object representation (main loop)
  // index is needed as we need to find also corresponding molecule info data row (row 0)
  for (std::vector<BnxRow>::size_type m = 0; m != (*bnxFile->_rows)[BnxRowType::Data0]->size(); m++)
  {
    // shortcut to access fields of molecule header in my test BNX file (not correct generally)
    // run id is last column and molecule id is first column 
    auto runId = stoi((*(*bnxFile->_rows)[BnxRowType::Data0])[m].RowData->back());
    auto moleculeId = stoi((*(*(*bnxFile->_rows)[BnxRowType::Data0])[m].RowData)[0]);
    
    // take molecule data row
    auto data = (*(*bnxFile->_rows)[BnxRowType::Data1])[m];
    auto dataRowSize = data.RowData->size();
    
    // go through the labels of the molecule and convert them to numbers and distances between labels
    // BNX("100", "1100", "2000") will be converted to RUN(1000, 900) as we are not interested in absolute postitions of labels
    // also we need to create reverse molecule here to account for possibility that molecule is scanned in reverse order 
    auto molecule = new Molecule<T>(moleculeId, false);
    for (std::vector<string>::size_type l = 0; l != dataRowSize - 1; l++)
    {
      auto val1 = stoul((*data.RowData)[l]);
      auto val2 = stoul((*data.RowData)[l+1]);
      auto diff = val2 - val1;

      molecule->_labelDistances.push_back(diff);
      molecule->_labelsCount++;
      
      if (molecule->_maxLabelDistance < diff)
      {
        molecule->_maxLabelDistance = diff;
      }

      _labelDistances[diff].Count++;
    }
    
    // generate reverse order molecule from the original
    Molecule<T>* reverseMolecule;
    if (generateReverseMolecules)
    {
      reverseMolecule = new Molecule<T>(moleculeId, true);
      for (int32_t idx = molecule->_labelDistances.size() - 1; idx >= 0; idx--)
      {
        reverseMolecule->_labelDistances.push_back(molecule->_labelDistances[idx]);
        reverseMolecule->_labelsCount++;
      }
    }

    // fill the run object only if count of transformed labels is greater than zero
    // molecules containing only one scanned labels are reduced to zero relative label distances (there is no distance to compute)
    // molecules must contain more than specified amount of transformed labels
    auto run = _data[runId];
    if (molecule->_labelsCount > 0 && 
        molecule->_labelsCount >= minMoleculeLabelsCnt)
    {
      run->_molecules.push_back(molecule);
      run->_moleculesCount++;

      if (generateReverseMolecules)
      {
        run->_molecules.push_back(reverseMolecule);
        run->_moleculesCount++;
      }

      if (run->_maxLabelDistance < molecule->_maxLabelDistance)
      {
        molecule->_maxLabelDistance = molecule->_maxLabelDistance;
      }

      if (_maxLabelDistance < molecule->_maxLabelDistance)
      {
        _maxLabelDistance = molecule->_maxLabelDistance;
      }
    }
    else
    {
      run->_leftOutMoleculesCount++;
    }
  }
  
  auto end = chrono::steady_clock::now();

  if (printStatistics)
  {
    chrono::duration<double> diff = end - start;
    cout << "\nTime to fill RunsData object: " << diff.count() << "s\n";

    // for now we are not interested in printing out all run statistics
    //PrintStatistics();
  }
}

template<typename T>
void RunsData<T>::PrintStatistics()
{
  cout << "\nRunsData Statistics\n\n";

  int cntOfDistinctLabels = 0;
  for (auto const& r : _labelDistances)
  {
    cntOfDistinctLabels++;
  }
  cout << "Distinct relative labels count: " << cntOfDistinctLabels << "\n\n";

  for (auto const& r : _data)
  {
    auto run = _data[r.first];

    if (run->MoleculesCount > 0)
    {
      cout << "****************** RUN " << r.first << " ******************\n";
      cout << "Number of molecules: " << run->MoleculesCount << "\n";
      cout << "Number of left out molecules: " << run->LeftOutMoleculesCount << "\n";

      // lets find longest and shortest molecule and avg and total length of molecules (in labels) 
      long minLabelsCnt = ULONG_MAX;
      long maxLabelsCnt = 0;
      long totalLabelsCnt = 0;
      
      for (auto const& m : (*run->Molecules))
      {
        totalLabelsCnt += m->LabelsCount;
        
        if (minLabelsCnt > m->LabelsCount)
        {
          minLabelsCnt = m->LabelsCount;

          if (m->LabelsCount == 0)
          {
            cout << "First molecule ID with no labels: " << m->MoleculeId << "\n";
          }
        }
        
        if (maxLabelsCnt < m->LabelsCount)
        {
          maxLabelsCnt = m->LabelsCount;
        }
      }

      cout << "Min cnt of labels in molecules: " << minLabelsCnt << "\n";
      cout << "Max cnt of labels in molecules: " << maxLabelsCnt << "\n";
      cout << "Tot cnt of labels in molecules: " << totalLabelsCnt << "\n"; 
      cout << "Avg cnt of labels in molecules: " << totalLabelsCnt / run->MoleculesCount << "\n\n";
    }
  }
}

#endif