#ifndef __TST_H__
#define __TST_H__

#include <immintrin.h>
#include <vector>
#include <chrono>
#include <iostream>

#include "defs.h"
#include "tst_arr.h"
#include "../Configuration/configuration.h"

// tuple for residual data - first vector contain keys, second vector contains values which belongs to the sequence of keys from the first vector
// bool specifies if key vector is finalized and no other key values can't be appended to it
using RESID_TUPLE_TYPE = std::tuple<std::vector<int32_t>*, std::vector<int32_t>*, bool>;

class tst
{
private:
    tst_arr* tst_arr_;
    int32_t node_stop_creation_depth_;

    // vector with keys (words) which should be inserted into the TST
    std::vector<std::vector<int32_t>>* keys_vect_;

    // pointer vectors
    std::vector<std::vector<RESID_TUPLE_TYPE*>*> resid_vect_ptrs_;
    std::vector<RESID_TUPLE_TYPE*> resid_ptrs_;
    std::vector<std::vector<int32_t>*> int_vect_ptrs_;
    
    // vector of values which matches indexes of keys which should be indexed at the end of the key sequence (word)
    std::vector<int32_t>* values_vect_;

    // residual data pointer array
    int64_t resid_ptr_arr_[8];
    std::vector<std::vector<int32_t>*> resid_vals_arr_[8];
    std::vector<int32_t> resid_ins_keys_[8];

    // statistics
    uint32_t words_cnt_;

    // REGISTERS
    // registers needed for work with words and letters in them
    __tst_256i WORD_IDX, WORD_LENGTHS, LETTER_IDX, LAST_LETTER, WORD_LONGER_THAN_ZERO, WORD_IDX_MAX, LETTER_IDX_GT_PARAM;

    __tst_256i RESID_VECT_IDX, RESID_FOUND;

    // KEYS contains letters for insertion
    __tst_256i KEYS;

    // insertion control registers
    __tst_256i CURR_NODE_IDX, NODE_KEYS, CMP_EQ, CMP_LT, CMP_GT;

    // number constat registers
    __tst_256i MINUS_ONE, ZERO, ONE, EIGHT, NODE_STOP_CREATION_LETTER_IDX;

    void init()
    {
        MINUS_ONE.m256i = _mm256_set1_epi32(-1);
        ZERO.m256i = _mm256_set1_epi32(0);
        ONE.m256i = _mm256_set1_epi32(1);
        EIGHT.m256i = _mm256_set1_epi32(8);

        NODE_KEYS = ZERO;
        CMP_LT = ZERO;
        CMP_GT = ZERO;
        CURR_NODE_IDX = ZERO;
        KEYS = ZERO;

        WORD_IDX.m256i = _mm256_set_epi32(-1, -2, -3, -4, -5, -6, -7, -8);
        LETTER_IDX = MINUS_ONE;
        WORD_LENGTHS = MINUS_ONE;
        WORD_LONGER_THAN_ZERO = ZERO;
        LAST_LETTER = MINUS_ONE;
        CMP_EQ = MINUS_ONE;
        NODE_STOP_CREATION_LETTER_IDX.m256i = _mm256_set1_epi32(node_stop_creation_depth_ - 1);

        RESID_VECT_IDX = ZERO;
    }

public:
    // storage for profiling times
    std::chrono::nanoseconds upd_val_reg_t, upd_node_val_reg_t, ins_val_t, cmp_t, upd_letter_t, ins_idx_t, upd_curr_idx_t, upd_word_reg_t, search_res_t;

    /// <summary>
    /// Constructor
    /// </summary>
    /// <param name="nodes_cnt"></param>
    /// <returns></returns>
    tst(uint32_t nodes_cnt, int32_t node_stop_creation_depth)  // NOLINT(cppcoreguidelines-pro-type-member-init)
    {
        words_cnt_ = 0;
        tst_arr_ = new tst_arr(nodes_cnt);
        tst_arr_->last_tst_idx_ = 0;
        node_stop_creation_depth_ = node_stop_creation_depth;
    }

    int32_t get_tst_size()
    {
        return tst_arr_->get_total_size_kb();
    }

    void clear()
    {
        auto size1 = resid_vect_ptrs_.size();
        for (auto ptr : resid_vect_ptrs_)
        {
            delete ptr;
        }

        auto size2 = resid_ptrs_.size();
        for (auto ptr : resid_ptrs_)
        {
            delete ptr;
        }

        auto size3 = int_vect_ptrs_.size();
        for (auto ptr : int_vect_ptrs_)
        {
            delete ptr;
        }

        delete tst_arr_;
    }

    /// <summary>
    /// Frees some memory after the TST is filled with data
    /// </summary>
    void clean_up_memory()
    {
        tst_arr_->clean_up_memory();
    }

    /// <summary>
    /// Inserts vector of words into the TST
    /// </summary>
    /// <param name="keys"></param>
    /// <param name="values"></param>
    /// <param name="keys_vect_size"></param>
    void insert(std::vector<std::vector<int32_t>>* keys, std::vector<int32_t>* values, int32_t keys_vect_size)
    {
        init();

        keys_vect_ = keys;
        values_vect_ = values;
        WORD_IDX_MAX.m256i = _mm256_set1_epi32(keys_vect_size - 1);

        init_word_reg();

        do
        {
            LETTER_IDX_GT_PARAM.m256i = _mm256_cmpgt_epi32(LETTER_IDX.m256i, NODE_STOP_CREATION_LETTER_IDX.m256i);

            update_keys_reg();
            update_node_keys_reg();
            insert_keys();
            update_node_keys_reg();
            insert_values();

            CMP_EQ.m256i = _mm256_cmpeq_epi32(KEYS.m256i, NODE_KEYS.m256i);
            CMP_LT.m256i = _mm256_cmpgt_epi32(NODE_KEYS.m256i, KEYS.m256i);
            CMP_GT.m256i = _mm256_cmpgt_epi32(KEYS.m256i, NODE_KEYS.m256i);

            update_letter_reg();
            insert_idx();
            update_word_reg();
        } while (_mm256_movemask_ps(_mm256_castsi256_ps(WORD_LONGER_THAN_ZERO.m256i)) > 0);
    }

    /// <summary>
    /// Tries to find words in the TST
    /// Returns vector of booleans if words has been found in the TST
    /// </summary>
    /// <param name="words"></param>
    /// <param name="words_size"></param>
    /// <param name="ret_val"></param>
    /// <returns></returns>
    void search(std::vector<std::vector<int32_t>>* words, int32_t words_size, std::vector<std::vector<int32_t>>& ret_val)
    {
        init();

        // init words storage
        keys_vect_ = words;
        WORD_IDX_MAX.m256i = _mm256_set1_epi32(words_size - 1);

        // init of words and letters
        init_word_reg();

        do
        {
            LETTER_IDX_GT_PARAM.m256i = _mm256_cmpgt_epi32(LETTER_IDX.m256i, NODE_STOP_CREATION_LETTER_IDX.m256i);

            // read keys from words and from nodes to registers for comparison purposes
            update_keys_reg();
            update_node_keys_reg();

            search_residual_data();

            // compare keys
            CMP_EQ.m256i = _mm256_cmpeq_epi32(KEYS.m256i, NODE_KEYS.m256i);
            CMP_LT.m256i = _mm256_cmpgt_epi32(NODE_KEYS.m256i, KEYS.m256i);
            CMP_GT.m256i = _mm256_cmpgt_epi32(KEYS.m256i, NODE_KEYS.m256i);

            // process positive search results
            for (int32_t i = 0; i <= 7; i++)
            {
                if (CMP_EQ.m256i_i32[i] == -1 && LAST_LETTER.m256i_i32[i] == -1 && tst_arr_->get_ew_bit(CURR_NODE_IDX.m256i_i32[i]))
                {
                    auto vals = reinterpret_cast<std::vector<int32_t>*>(tst_arr_->get_values_ptr(CURR_NODE_IDX.m256i_i32[i]));

                    for (auto& val : (*vals))
                    {
                        ret_val[WORD_IDX.m256i_i32[i]].push_back(val);
                    }
                }

                if (RESID_FOUND.m256i_i32[i] == -1)
                {
                    for (auto& val_vect : resid_vals_arr_[i])
                    {
                        for (auto& val : (*val_vect))
                        {
                            ret_val[WORD_IDX.m256i_i32[i]].push_back(val);
                        }
                    }
                }
            }

            update_letter_reg();
            update_curr_idx();
            update_word_reg_search();
        } while (_mm256_movemask_ps(_mm256_castsi256_ps(WORD_LONGER_THAN_ZERO.m256i)) > 0);
    }

    /// <summary>
    /// Searches residual data for key
    /// </summary>
    inline void search_residual_data()
    {
        RESID_FOUND = ZERO;

        for (int i = 0; i <= 7; i++)
        {
            if (LETTER_IDX_GT_PARAM.m256i_i32[i] != -1)
            {
                continue;
            }

            // remember residual vector pointer and set residual vector index to zero => start of comparisons
            if (LETTER_IDX.m256i_i32[i] - 1 == NODE_STOP_CREATION_LETTER_IDX.m256i_i32[i])
            {
                auto* resid_ptr = reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(tst_arr_->get_resid_ptr(CURR_NODE_IDX.m256i_i32[i]));
                resid_ptr_arr_[i] = reinterpret_cast<int64_t>(resid_ptr);

                RESID_VECT_IDX.m256i_i32[i] = 0;
            }

            auto* resid_ptr = reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(resid_ptr_arr_[i]);
            
            // there are no residual data vector on the node => key can't be found
            if (resid_ptr == reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(0xFFFFFFFFFFFFFFFF))
            {
                continue;
            }

            // search all residual data key vectors for the key
            resid_vals_arr_[i].clear();
            bool increment = false;
            for (auto& resid_it : *resid_ptr)
            {
                auto* resid_key_vector = std::get<0>(*resid_it);

                if (resid_key_vector->size() >= RESID_VECT_IDX.m256i_i32[i] && (*resid_key_vector)[RESID_VECT_IDX.m256i_i32[i]] == KEYS.m256i_i32[i])
                {
                    if (LAST_LETTER.m256i_i32[i] == -1 && resid_key_vector->size() - 1 == RESID_VECT_IDX.m256i_i32[i])
                    {
                        resid_vals_arr_[i].push_back(std::get<1>(*resid_it));
                        RESID_FOUND.m256i_i32[i] = -1;
                    }

                    increment = true;
                }
            }

            if (increment)
            {
                RESID_VECT_IDX.m256i_i32[i]++;
            }
        }
    }

    /// <summary>
    /// Updates word register
    /// </summary>
    inline void update_word_reg_search()
    {
        __tst_256i word_idx_eq_minus_one;
        word_idx_eq_minus_one.m256i = _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i);
        __tst_256i word_idx_plus_8;
        word_idx_plus_8.m256i = _mm256_add_epi32(WORD_IDX.m256i, EIGHT.m256i);
        word_idx_plus_8.m256i = _mm256_blendv_epi8(word_idx_plus_8.m256i, MINUS_ONE.m256i, word_idx_eq_minus_one.m256i);

        __tst_256i word_gt_len;
        word_gt_len.m256i = _mm256_cmpgt_epi32(word_idx_plus_8.m256i, WORD_IDX_MAX.m256i);
        __tst_256i increment_word;
        increment_word.m256i = _mm256_cmpeq_epi32(LETTER_IDX.m256i, MINUS_ONE.m256i);
        
        // increment word index also when there is no way to continue in current word (so when CURR_IDX == -1)
        // but do not increment word if we are searching through residual data
        __tst_256i cond;
        cond.m256i = _mm256_cmpeq_epi32(CURR_NODE_IDX.m256i, MINUS_ONE.m256i);
        cond.m256i = _mm256_blendv_epi8(cond.m256i, ZERO.m256i, LETTER_IDX_GT_PARAM.m256i);

        increment_word.m256i = _mm256_blendv_epi8(increment_word.m256i, MINUS_ONE.m256i, cond.m256i);

        __tst_256i new_word_idx;
        new_word_idx.m256i = _mm256_blendv_epi8(WORD_IDX.m256i, word_idx_plus_8.m256i, increment_word.m256i);
        new_word_idx.m256i = _mm256_blendv_epi8(new_word_idx.m256i, MINUS_ONE.m256i, word_gt_len.m256i);

        WORD_IDX.m256i = _mm256_blendv_epi8(WORD_IDX.m256i, new_word_idx.m256i, increment_word.m256i);
        CURR_NODE_IDX.m256i = _mm256_blendv_epi8(CURR_NODE_IDX.m256i, ZERO.m256i, increment_word.m256i);
        CURR_NODE_IDX.m256i = _mm256_blendv_epi8(CURR_NODE_IDX.m256i, MINUS_ONE.m256i, _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i));
        LETTER_IDX.m256i = _mm256_blendv_epi8(LETTER_IDX.m256i, ZERO.m256i, increment_word.m256i);

        WORD_LENGTHS.m256i = _mm256_set_epi32
        (
            WORD_IDX.m256i_i32[7] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[7]].size()),
            WORD_IDX.m256i_i32[6] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[6]].size()),
            WORD_IDX.m256i_i32[5] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[5]].size()),
            WORD_IDX.m256i_i32[4] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[4]].size()),
            WORD_IDX.m256i_i32[3] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[3]].size()),
            WORD_IDX.m256i_i32[2] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[2]].size()),
            WORD_IDX.m256i_i32[1] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[1]].size()),
            WORD_IDX.m256i_i32[0] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[0]].size())
        );

        WORD_LONGER_THAN_ZERO.m256i = _mm256_cmpgt_epi32(WORD_LENGTHS.m256i, ZERO.m256i);
        WORD_LONGER_THAN_ZERO.m256i = _mm256_blendv_epi8(WORD_LONGER_THAN_ZERO.m256i, ZERO.m256i, _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i));

        __tst_256i letter_idx_plus_one;
        letter_idx_plus_one.m256i = _mm256_add_epi32(LETTER_IDX.m256i, ONE.m256i);
        LAST_LETTER.m256i = _mm256_cmpeq_epi32(letter_idx_plus_one.m256i, WORD_LENGTHS.m256i);
    }

    /// <summary>
    /// Inserts keys
    /// </summary>
    inline void insert_keys()
    {
#pragma unroll
        for (int16_t i = 0; i <= 7; i++)
        {
            if (CURR_NODE_IDX.m256i_i32[i] == -1)
            {
                continue;
            }

            // if letter index is greater than set parameter then store keys in array
            if (LETTER_IDX_GT_PARAM.m256i_i32[i] == -1)
            {
                auto* resid_main_vector = reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(tst_arr_->get_resid_ptr(CURR_NODE_IDX.m256i_i32[i]));

                // init vector which holds all residual data objects - main resid data vector on the node
                if (LETTER_IDX.m256i_i32[i] - 1 == NODE_STOP_CREATION_LETTER_IDX.m256i_i32[i])
                {
                    if (resid_main_vector == reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(0xFFFFFFFFFFFFFFFF))
                    {
                        resid_main_vector = new std::vector<RESID_TUPLE_TYPE*>();
                        tst_arr_->set_resid_ptr(CURR_NODE_IDX.m256i_i32[i], resid_main_vector);
                        resid_vect_ptrs_.push_back(resid_main_vector);  // garbage collection
                    }
                }

                bool resid_init_needed = true;
                bool found = false;

                // store key into the vector so we have whole key sequence for lookup / insert
                resid_ins_keys_[i].push_back(KEYS.m256i_i32[i]);

                // try to look up residual data structure which corresponds to key which we are trying to insert
                if (resid_main_vector != reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(0xFFFFFFFFFFFFFFFF) && resid_main_vector->size() != 0)
                {
                    // go through all residual data tuples in the vector
                    for (auto& resid_iter : (*resid_main_vector))
                    {
                        for (int32_t key_idx = 0; key_idx < resid_ins_keys_[i].size(); key_idx++)
                        {
                            // new key is longer than the key stored in this particular residual data object
                            if (key_idx >= std::get<0>(*resid_iter)->size()) 
                            {
                                found = false;
                                break;
                            }

                            // new key is different from the stored key
                            if ((*std::get<0>(*resid_iter))[key_idx] != resid_ins_keys_[i][key_idx])
                            {
                                resid_init_needed = true;
                                found = false;
                                break;
                            }
                            // keys are equal
                            else
                            {
                                // resid object must be initialized always when resid object is locked and we are on the last letter
                                resid_init_needed = std::get<2>(*resid_iter) && LAST_LETTER.m256i_i32[i] == -1;

                                // if sizes are equal => do not initialize new resid object
                                if (resid_ins_keys_[i].size() == std::get<0>(*resid_iter)->size())
                                {
                                    resid_init_needed = false;
                                }

                                found = true;
                            }
                        }

                        // if everything matches then store pointer in the array and break out from the loop
                        if (found && !resid_init_needed)
                        {
                            resid_ptr_arr_[i] = reinterpret_cast<int64_t>(resid_iter);
                            break;
                        }
                    }
                }
                
                // init residual data object if resid main vector is empty or resid storage was not found
                // of if this resid data object is locked for adding
                if (resid_init_needed)
                {
                    auto* resid = new RESID_TUPLE_TYPE();
                    std::get<0>((*resid)) = new std::vector<int32_t>();
                    std::get<1>((*resid)) = new std::vector<int32_t>();
                    std::get<2>((*resid)) = false;

                    resid_main_vector->push_back(resid);
                    resid_ptr_arr_[i] = reinterpret_cast<int64_t>(resid);

                    // when initializing always copy all keys into the resid object
                    for (auto& key : resid_ins_keys_[i])
                    {
                        std::get<0>((*resid))->push_back(key);
                    }
                    
                    resid_ptrs_.push_back(resid);                       // garbage collection
                    int_vect_ptrs_.push_back(std::get<0>((*resid)));    // garbage collection
                    int_vect_ptrs_.push_back(std::get<1>((*resid)));    // garbage collection
                }
                else
                {
                    if (!found)
                    {
                        auto* resid = reinterpret_cast<RESID_TUPLE_TYPE*>(resid_ptr_arr_[i]);
                        std::get<0>((*resid))->push_back(KEYS.m256i_i32[i]);
                    }
                }

                // clear vector when last letter of the word was processed and lock resid object for further modification
                if (LAST_LETTER.m256i_i32[i] == -1)
                {
                    auto* resid = reinterpret_cast<RESID_TUPLE_TYPE*>(resid_ptr_arr_[i]);
                    std::get<2>(*resid) = true;
                    resid_ins_keys_[i].clear();
                }
            }
            // otherwise store keys in TST nodes
            else
            {
                __tst_256i node_empty;
                node_empty.m256i = _mm256_cmpeq_epi32(MINUS_ONE.m256i, NODE_KEYS.m256i);

                if (node_empty.m256i_i32[i] == -1)
                {
                    tst_arr_->set_key(CURR_NODE_IDX.m256i_i32[i], KEYS.m256i_i32[i]);
                }
            }
        }
    }

    /// <summary>
    /// Inserts values
    /// </summary>
    inline void insert_values()
    {
#pragma unroll
        for (int16_t i = 0; i <= 7; i++)
        {
            // if letter index is greater than set parameter then store keys in array
            if (LETTER_IDX_GT_PARAM.m256i_i32[i] == -1 && LAST_LETTER.m256i_i32[i] == -1)
            {
                auto* resid = reinterpret_cast<RESID_TUPLE_TYPE*>(resid_ptr_arr_[i]);
                auto val  = (*values_vect_)[WORD_IDX.m256i_i32[i]];
                std::get<1>((*resid))->push_back(val);
            }
            // otherwise store keys in TST nodes
            else
            {
                if (KEYS.m256i_i32[i] != NODE_KEYS.m256i_i32[i] || (KEYS.m256i_i32[i] == -1 && NODE_KEYS.m256i_i32[i] == -1))
                {
                    continue;
                }

                if (LAST_LETTER.m256i_i32[i] == -1)
                {
                    tst_arr_->set_ew_bit(CURR_NODE_IDX.m256i_i32[i]);

                    std::vector<int32_t>* val_ptr = reinterpret_cast<std::vector<int32_t>*>(tst_arr_->get_values_ptr(CURR_NODE_IDX.m256i_i32[i]));
                    if (val_ptr == reinterpret_cast<std::vector<int32_t>*>(0xFFFFFFFFFFFFFFFF))
                    {
                        val_ptr = new std::vector<int32_t>();
                        int_vect_ptrs_.push_back(val_ptr);
                        tst_arr_->set_values_ptr(CURR_NODE_IDX.m256i_i32[i], val_ptr);
                    }

                    val_ptr->push_back((*values_vect_)[WORD_IDX.m256i_i32[i]]);
                    words_cnt_++;
                }
            }
        }
    }

    /// <summary>
    /// Function inserts EQ, LT and GT node indexes into the currently processed nodes based on CMP_XX registers
    /// </summary>
    inline void insert_idx()
    {
#pragma unroll
        for (int16_t i = 0; i <= 7; i++)
        {
            if (CURR_NODE_IDX.m256i_i32[i] == -1 || LETTER_IDX.m256i_i32[i] == -1 || LETTER_IDX_GT_PARAM.m256i_i32[i] == -1)
            {
                continue;
            }

            if (tst_arr_->get_lt_idx_cmp(CURR_NODE_IDX.m256i_i32[i], CMP_LT.m256i_i32[i]))
            {
                tst_arr_->set_lt_idx(CURR_NODE_IDX.m256i_i32[i], ++tst_arr_->last_tst_idx_);
            }
            else if (tst_arr_->get_eq_idx_cmp(CURR_NODE_IDX.m256i_i32[i], CMP_EQ.m256i_i32[i]))
            {
                tst_arr_->set_eq_idx(CURR_NODE_IDX.m256i_i32[i], ++tst_arr_->last_tst_idx_);
            }
            else if (tst_arr_->get_gt_idx_cmp(CURR_NODE_IDX.m256i_i32[i], CMP_GT.m256i_i32[i]))
            {
                tst_arr_->set_gt_idx(CURR_NODE_IDX.m256i_i32[i], ++tst_arr_->last_tst_idx_);
            }

            if (CURR_NODE_IDX.m256i_i32[i] == -1)
            {
                continue;
            }

            if (CMP_LT.m256i_i32[i] == -1)
            {
                CURR_NODE_IDX.m256i_i32[i] = tst_arr_->get_lt_idx(CURR_NODE_IDX.m256i_i32[i]);
            }
            else if (CMP_EQ.m256i_i32[i] == -1)
            {
                CURR_NODE_IDX.m256i_i32[i] = tst_arr_->get_eq_idx(CURR_NODE_IDX.m256i_i32[i]);
            }
            else if (CMP_GT.m256i_i32[i] == -1)
            {
                CURR_NODE_IDX.m256i_i32[i] = tst_arr_->get_gt_idx(CURR_NODE_IDX.m256i_i32[i]);
            }
        }
    }

    /// <summary>
    /// Function moves CURR_IDX to correct nodes based on CMP_XX registers
    /// </summary>
    inline void update_curr_idx()
    {
#pragma unroll
        for (int16_t i = 0; i <= 7; i++)
        {
            if (CURR_NODE_IDX.m256i_i32[i] == -1)
            {
                continue;
            }

            if (CMP_LT.m256i_i32[i] == -1)
            {
                CURR_NODE_IDX.m256i_i32[i] = tst_arr_->get_lt_idx(CURR_NODE_IDX.m256i_i32[i]);
            }
            else if (CMP_EQ.m256i_i32[i] == -1)
            {
                CURR_NODE_IDX.m256i_i32[i] = tst_arr_->get_eq_idx(CURR_NODE_IDX.m256i_i32[i]);
            }
            else if (CMP_GT.m256i_i32[i] == -1)
            {
                CURR_NODE_IDX.m256i_i32[i] = tst_arr_->get_gt_idx(CURR_NODE_IDX.m256i_i32[i]);
            }
        }
    }

    /// <summary>
    /// Initializes word registers when insert starts
    /// </summary>
    inline void init_word_reg()
    {
        __tst_256i word_idx_plus_8;
        word_idx_plus_8.m256i = _mm256_add_epi32(WORD_IDX.m256i, EIGHT.m256i);
        __tst_256i word_gt_len;
        word_gt_len.m256i = _mm256_cmpgt_epi32(word_idx_plus_8.m256i, WORD_IDX_MAX.m256i);
        __tst_256i increment_word;
        increment_word.m256i = _mm256_cmpeq_epi32(LETTER_IDX.m256i, MINUS_ONE.m256i);

        __tst_256i new_word_idx;
        new_word_idx.m256i = _mm256_blendv_epi8(WORD_IDX.m256i, word_idx_plus_8.m256i, increment_word.m256i);
        new_word_idx.m256i = _mm256_blendv_epi8(new_word_idx.m256i, MINUS_ONE.m256i, word_gt_len.m256i);

        WORD_IDX.m256i = _mm256_blendv_epi8(WORD_IDX.m256i, new_word_idx.m256i, increment_word.m256i);
        CURR_NODE_IDX.m256i = _mm256_blendv_epi8(CURR_NODE_IDX.m256i, ZERO.m256i, increment_word.m256i);
        CURR_NODE_IDX.m256i = _mm256_blendv_epi8(CURR_NODE_IDX.m256i, MINUS_ONE.m256i, _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i));
        LETTER_IDX.m256i = _mm256_blendv_epi8(LETTER_IDX.m256i, ZERO.m256i, increment_word.m256i);

        WORD_LENGTHS.m256i = _mm256_set_epi32
        (
            WORD_IDX.m256i_i32[7] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[7]].size()),
            WORD_IDX.m256i_i32[6] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[6]].size()),
            WORD_IDX.m256i_i32[5] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[5]].size()),
            WORD_IDX.m256i_i32[4] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[4]].size()),
            WORD_IDX.m256i_i32[3] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[3]].size()),
            WORD_IDX.m256i_i32[2] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[2]].size()),
            WORD_IDX.m256i_i32[1] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[1]].size()),
            WORD_IDX.m256i_i32[0] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[0]].size())
        );

        WORD_LONGER_THAN_ZERO.m256i = _mm256_cmpgt_epi32(WORD_LENGTHS.m256i, ZERO.m256i);

        __tst_256i letter_idx_plus_one;
        letter_idx_plus_one.m256i = _mm256_add_epi32(LETTER_IDX.m256i, ONE.m256i);
        LAST_LETTER.m256i = _mm256_cmpeq_epi32(letter_idx_plus_one.m256i, WORD_LENGTHS.m256i);
    }

    /// <summary>
    /// Updates word register
    /// </summary>
    inline void update_word_reg()
    {
        __tst_256i word_idx_eq_minus_one;
        word_idx_eq_minus_one.m256i = _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i);
        __tst_256i word_idx_plus_8;
        word_idx_plus_8.m256i = _mm256_add_epi32(WORD_IDX.m256i, EIGHT.m256i);
        word_idx_plus_8.m256i = _mm256_blendv_epi8(word_idx_plus_8.m256i, MINUS_ONE.m256i, word_idx_eq_minus_one.m256i);

        __tst_256i word_gt_len;
        word_gt_len.m256i = _mm256_cmpgt_epi32(word_idx_plus_8.m256i, WORD_IDX_MAX.m256i);
        __tst_256i increment_word;
        increment_word.m256i = _mm256_cmpeq_epi32(LETTER_IDX.m256i, MINUS_ONE.m256i);

        __tst_256i new_word_idx;
        new_word_idx.m256i = _mm256_blendv_epi8(WORD_IDX.m256i, word_idx_plus_8.m256i, increment_word.m256i);
        new_word_idx.m256i = _mm256_blendv_epi8(new_word_idx.m256i, MINUS_ONE.m256i, word_gt_len.m256i);

        WORD_IDX.m256i = _mm256_blendv_epi8(WORD_IDX.m256i, new_word_idx.m256i, increment_word.m256i);
        CURR_NODE_IDX.m256i = _mm256_blendv_epi8(CURR_NODE_IDX.m256i, ZERO.m256i, increment_word.m256i);
        CURR_NODE_IDX.m256i = _mm256_blendv_epi8(CURR_NODE_IDX.m256i, MINUS_ONE.m256i, _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i));
        LETTER_IDX.m256i = _mm256_blendv_epi8(LETTER_IDX.m256i, ZERO.m256i, increment_word.m256i);

        WORD_LENGTHS.m256i = _mm256_set_epi32
        (
            WORD_IDX.m256i_i32[7] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[7]].size()),
            WORD_IDX.m256i_i32[6] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[6]].size()),
            WORD_IDX.m256i_i32[5] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[5]].size()),
            WORD_IDX.m256i_i32[4] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[4]].size()),
            WORD_IDX.m256i_i32[3] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[3]].size()),
            WORD_IDX.m256i_i32[2] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[2]].size()),
            WORD_IDX.m256i_i32[1] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[1]].size()),
            WORD_IDX.m256i_i32[0] == -1 ? -1 : static_cast<int>((*keys_vect_)[WORD_IDX.m256i_i32[0]].size())
        );

        WORD_LONGER_THAN_ZERO.m256i = _mm256_cmpgt_epi32(WORD_LENGTHS.m256i, ZERO.m256i);
        WORD_LONGER_THAN_ZERO.m256i = _mm256_blendv_epi8(WORD_LONGER_THAN_ZERO.m256i, ZERO.m256i, _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i));

        __tst_256i letter_idx_plus_one;
        letter_idx_plus_one.m256i = _mm256_add_epi32(LETTER_IDX.m256i, ONE.m256i);
        LAST_LETTER.m256i = _mm256_cmpeq_epi32(letter_idx_plus_one.m256i, WORD_LENGTHS.m256i);
    }

    /// <summary>
    /// Move LETTER_IDX forward where CMP_EQ == -1
    /// </summary>
    inline void update_letter_reg()
    {
        __tst_256i letter_idx_plus_one;
        letter_idx_plus_one.m256i = _mm256_add_epi32(LETTER_IDX.m256i, ONE.m256i);
        __tst_256i gt;
        gt.m256i = _mm256_cmpgt_epi32(WORD_LENGTHS.m256i, letter_idx_plus_one.m256i);
        __tst_256i inner;
        inner.m256i = _mm256_blendv_epi8(MINUS_ONE.m256i, letter_idx_plus_one.m256i, gt.m256i);
        __tst_256i mask;
        mask.m256i = _mm256_blendv_epi8(CMP_EQ.m256i, MINUS_ONE.m256i, LETTER_IDX_GT_PARAM.m256i);

        LETTER_IDX.m256i = _mm256_blendv_epi8(LETTER_IDX.m256i, inner.m256i, mask.m256i);
    }

    /// <summary>
    /// Fills KEYS register with letters from input vector if WORD_IDX and LETTER_IDX registers contain valid values
    /// </summary>
    inline void update_keys_reg()
    {
        __tst_256i do_not_update_by_word;
        do_not_update_by_word.m256i = _mm256_cmpeq_epi32(WORD_IDX.m256i, MINUS_ONE.m256i);
        __tst_256i do_not_update_by_letter;
        do_not_update_by_letter.m256i = _mm256_cmpeq_epi32(LETTER_IDX.m256i, MINUS_ONE.m256i);
        __tst_256i do_not_update;
        do_not_update.m256i = _mm256_or_si256(do_not_update_by_word.m256i, do_not_update_by_letter.m256i);
        
        KEYS.m256i = _mm256_set_epi32
        (
            do_not_update.m256i_i32[7] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[7]][LETTER_IDX.m256i_i32[7]],
            do_not_update.m256i_i32[6] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[6]][LETTER_IDX.m256i_i32[6]],
            do_not_update.m256i_i32[5] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[5]][LETTER_IDX.m256i_i32[5]],
            do_not_update.m256i_i32[4] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[4]][LETTER_IDX.m256i_i32[4]],
            do_not_update.m256i_i32[3] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[3]][LETTER_IDX.m256i_i32[3]],
            do_not_update.m256i_i32[2] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[2]][LETTER_IDX.m256i_i32[2]],
            do_not_update.m256i_i32[1] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[1]][LETTER_IDX.m256i_i32[1]],
            do_not_update.m256i_i32[0] == -1 ? -1 : (*keys_vect_)[WORD_IDX.m256i_i32[0]][LETTER_IDX.m256i_i32[0]]
        );
    }

    /// <summary>
    /// Function reads values from TST nodes according to CURR_IDX register and stores them in NODE_KEYS register for comparisons
    /// </summary>
    inline void update_node_keys_reg()
    {
        __tst_256i curr_node_idx_eq_minus_one;
        curr_node_idx_eq_minus_one.m256i = _mm256_cmpeq_epi32(CURR_NODE_IDX.m256i, MINUS_ONE.m256i);
        __tst_256i cond;
        cond.m256i = _mm256_or_si256(curr_node_idx_eq_minus_one.m256i, LETTER_IDX_GT_PARAM.m256i);

        NODE_KEYS.m256i = _mm256_set_epi32(
            cond.m256i_i32[7] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[7]),
            cond.m256i_i32[6] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[6]),
            cond.m256i_i32[5] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[5]),
            cond.m256i_i32[4] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[4]),
            cond.m256i_i32[3] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[3]),
            cond.m256i_i32[2] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[2]),
            cond.m256i_i32[1] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[1]),
            cond.m256i_i32[0] == -1 ? -1 : tst_arr_->get_key(CURR_NODE_IDX.m256i_i32[0])
        );
    }

    /// <summary>
    /// Outputs whole TST to console
    /// </summary>
    /// <param name="node_idx"></param>
    void print_tst(const int32_t node_idx) const
    {
        if (node_idx == 0)
        {
            std::cout << "\n---------------------------------------\n";
        }

        if (tst_arr_->get_lt_idx(node_idx) > 0)
        {
            print_tst(tst_arr_->get_lt_idx(node_idx));
        }

        std::vector<int32_t>* val_ptr = reinterpret_cast<std::vector<int32_t>*>(tst_arr_->get_values_ptr(node_idx));

        std::cout << "IDX = " << node_idx << "\tKey = " << tst_arr_->get_key(node_idx);

        if (val_ptr == reinterpret_cast<std::vector<int32_t>*>(0xFFFFFFFFFFFFFFFF))
        {
            std::cout << "\tNo values";
        }
        else
        {
            std::cout << "\tVals = ";
            for (int j = 0; j < val_ptr->size(); ++j)
            {
                std::cout << (*val_ptr)[j] << ",";
            }
        }
            
        std::cout << 
            "\tEW = " << tst_arr_->get_ew_bit(node_idx) <<
            "\tLT = " << tst_arr_->get_lt_idx(node_idx) << 
            "\tEQ = " << tst_arr_->get_eq_idx(node_idx) << 
            "\tGT = " << tst_arr_->get_gt_idx(node_idx) << 
            "\tRESID = " << tst_arr_->get_resid_ptr(node_idx);

        // print out residual data
        auto* resid_ptr = reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(tst_arr_->get_resid_ptr(node_idx));
        if (resid_ptr != reinterpret_cast<std::vector<RESID_TUPLE_TYPE*>*>(0xFFFFFFFFFFFFFFFF))
        {
            for (auto& resid_it : (*resid_ptr))
            {
                std::cout << "\n -- KEYS -- \n";
                
                // keys
                for (auto keys_it : (*std::get<0>(*resid_it)))
                {
                    std::cout << keys_it << ",";
                }

                std::cout << "\n -- VALUES -- \n";

                // values
                for (auto vals_it : (*std::get<1>(*resid_it)))
                {
                    std::cout << vals_it << ",";
                }
            }
        }
        
        std::cout << "\n";

        if (tst_arr_->get_eq_idx(node_idx) > 0)
        {
            print_tst(tst_arr_->get_eq_idx(node_idx));
        }

        if (tst_arr_->get_gt_idx(node_idx) > 0)
        {
            print_tst(tst_arr_->get_gt_idx(node_idx));
        }
    }

    /// <summary>
    /// Output node to console
    /// </summary>
    /// <param name="node_idx"></param>
    void print_node(const int32_t node_idx) const
    {
        std::cout << "\n---------------------------------------";
        std::cout << "\nNODE_IDX : " << node_idx;
        std::cout << "\nVALUE    : " << tst_arr_->get_key(node_idx);
        std::cout << "\nLT       : " << tst_arr_->get_lt_idx(node_idx);
        std::cout << "\nEQ       : " << tst_arr_->get_eq_idx(node_idx);
        std::cout << "\nGT       : " << tst_arr_->get_gt_idx(node_idx);
        std::cout << "\nEW       : " << tst_arr_->get_ew_bit(node_idx);
    }

    /// <summary>
    /// Outputs TST stats to console
    /// </summary>
    void print_statistics() const
    {
        std::cout << "\n---------------------------------------";
        std::cout << "\nLAST_TST_INDEX : " << tst_arr_->last_tst_idx_;
        std::cout << "\nWords count    : " << words_cnt_;
    }

    uint64_t get_last_node_idx()
    {
        return tst_arr_->last_tst_idx_;
    }
};

#endif