#ifndef __TST_ARR_H__
#define __TST_ARR_H__

#include <iostream>
#include "../test_utils.h"

/// <summary>
///
/// Class which stores TST in array of bytes
/// Each node is represented as sequence of N bytes starting from some array index
/// Node consists of:
///     4 bytes - KEY
///     3x4 bytes - LEFT, EQ, RIGHT (child nodes indexes)
///     1x8 bytes - RESIDUAL DATA (pointer)
///     1x8 bytes - VALUES (pointer)
///
/// Metadata are stored in separate array - metadata_arr_
///
/// Residual data are stored as pointer to vector<vector<tuple<int32_t, int32_t>>>
/// Tuple stores key as first value and second value are data (for example molecule id)
///
/// </summary>
constexpr auto NODE_KEY_LENGTH = 4;
constexpr auto NODE_CHILD_IDX_LENGTH = 4;
constexpr auto NODE_RESID_LENGTH = 8;
constexpr auto NODE_VALUES_LENGTH = 8;

constexpr auto NODE_KEY_OFFSET = 0;
constexpr auto NODE_CHILD_LT_OFFSET = NODE_KEY_OFFSET + NODE_KEY_LENGTH;
constexpr auto NODE_CHILD_EQ_OFFSET = NODE_CHILD_LT_OFFSET + NODE_CHILD_IDX_LENGTH;
constexpr auto NODE_CHILD_GT_OFFSET = NODE_CHILD_EQ_OFFSET + NODE_CHILD_IDX_LENGTH;
constexpr auto NODE_RESID_OFFSET = NODE_CHILD_GT_OFFSET + NODE_CHILD_IDX_LENGTH;
constexpr auto NODE_VALUES_OFFSET = NODE_RESID_OFFSET + NODE_RESID_LENGTH;

constexpr auto NODE_BYTE_CNT = NODE_KEY_LENGTH + 3 * NODE_CHILD_IDX_LENGTH + NODE_RESID_LENGTH + NODE_VALUES_LENGTH;

constexpr auto METADATA_OFFSET = 0;
constexpr auto METADATA_BYTE_CNT = 1;
constexpr auto METADATA_EW_BIT = 0;

class tst_arr
{
private:
    unsigned char* tst_arr_;
    unsigned char* metadata_arr_;

    inline void set_arr_val_32(int32_t idx, const int32_t value, const int32_t offset) const
    {
        idx = idx * NODE_BYTE_CNT + offset;

        // be aware of endianity of different platforms
        memcpy(&tst_arr_[idx], &value, sizeof(int32_t));
        
        //idx = idx * NODE_BYTE_CNT_32 + offset;
        //reinterpret_cast<int32_t*>(tst_arr_)[idx] = value;
    }

    inline int32_t get_arr_val_32(int32_t idx, const int32_t offset) const
    {
        idx = idx * NODE_BYTE_CNT + offset;
        
        // be aware of endianity of different platforms
        int32_t ret_val = 0;
        memcpy(&ret_val, &tst_arr_[idx], sizeof(int32_t));
        return ret_val;

        /*idx = idx * NODE_BYTE_CNT_32 + offset;
        return reinterpret_cast<int32_t*>(tst_arr_)[idx];*/
    }

    inline void set_arr_val_64(int32_t idx, const int64_t value, const int32_t offset) const
    {
        if (idx < 0)
        {
            throw std::exception();
        }

        idx = idx * NODE_BYTE_CNT + offset;

        // be aware of endianity of different platforms
        memcpy(&tst_arr_[idx], &value, sizeof(int64_t));
    }

    inline int64_t get_arr_val_64(int32_t idx, const int32_t offset) const
    {
        if (idx < 0)
        {
            throw std::exception();
        }

        idx = idx * NODE_BYTE_CNT + offset;

        // be aware of endianity of different platforms
        int64_t ret_val = 0;
        memcpy(&ret_val, &tst_arr_[idx], sizeof(int64_t));
        return ret_val;
    }

public:
    // index of last used node in TST
    uint32_t last_tst_idx_ = 0;

    inline void clean_up_memory()
    {
        // tree array
        unsigned char* new_arr = new unsigned char[(last_tst_idx_ + 1) * NODE_BYTE_CNT];

        memcpy(new_arr, tst_arr_, (last_tst_idx_ + 1) * NODE_BYTE_CNT);

        delete[] tst_arr_;
        tst_arr_ = new_arr;

        // metadata array
        new_arr = new unsigned char[(last_tst_idx_ + 1) * METADATA_BYTE_CNT];

        memcpy(new_arr, metadata_arr_, (last_tst_idx_ + 1) * METADATA_BYTE_CNT);
        
        delete[] metadata_arr_;
        metadata_arr_ = new_arr;
    }

    /// <summary>
    /// Returns size of the internal arrays in kB
    /// </summary>
    /// <returns></returns>
    inline int32_t get_total_size_kb()
    {
        auto total_size = (last_tst_idx_ * (NODE_BYTE_CNT + METADATA_BYTE_CNT)) / 1024;
        return total_size;
    }

    /// <summary>
    /// Constructor
    /// </summary>
    /// <param name="nodes_cnt"></param>
    /// <returns></returns>
    explicit tst_arr(const int64_t nodes_cnt)
    {
        tst_arr_ = new unsigned char[NODE_BYTE_CNT * nodes_cnt];
        std::uninitialized_fill_n(tst_arr_, NODE_BYTE_CNT * nodes_cnt, -1);

        metadata_arr_ = new unsigned char[METADATA_BYTE_CNT * nodes_cnt];
        std::uninitialized_fill_n(metadata_arr_, METADATA_BYTE_CNT * nodes_cnt, 0);
    }

    inline void set_key(const int32_t node_idx, const int32_t value) const
    {
        set_arr_val_32(node_idx, value, NODE_KEY_OFFSET);
    }

    inline int32_t get_key(const int32_t node_idx) const
    {
        return get_arr_val_32(node_idx, NODE_KEY_OFFSET);
    }

    inline void set_lt_idx(const int32_t node_idx, const int32_t value) const
    {
        set_arr_val_32(node_idx, value, NODE_CHILD_LT_OFFSET);
    }

    inline int32_t get_lt_idx(const int32_t node_idx) const
    {
        return get_arr_val_32(node_idx, NODE_CHILD_LT_OFFSET);
    }

    inline bool get_lt_idx_cmp(const int32_t node_idx, const int32_t cmp) const
    {
        return get_arr_val_32(node_idx, NODE_CHILD_LT_OFFSET) == -1 && cmp == -1;
    }

    inline void set_eq_idx(const int32_t node_idx, const int32_t value) const
    {
        set_arr_val_32(node_idx, value, NODE_CHILD_EQ_OFFSET);
    }

    inline int32_t get_eq_idx(const int32_t node_idx) const
    {
        return get_arr_val_32(node_idx, NODE_CHILD_EQ_OFFSET);
    }

    inline bool get_eq_idx_cmp(const int32_t node_idx, const int32_t cmp) const
    {
        return get_arr_val_32(node_idx, NODE_CHILD_EQ_OFFSET) == -1 && cmp == -1;
    }

    inline void set_gt_idx(const int32_t node_idx, const int32_t value) const
    {
        set_arr_val_32(node_idx, value, NODE_CHILD_GT_OFFSET);
    }

    inline int32_t get_gt_idx(const int32_t node_idx) const
    {
        return get_arr_val_32(node_idx, NODE_CHILD_GT_OFFSET);
    }

    inline bool get_gt_idx_cmp(const int32_t node_idx, const int32_t cmp) const
    {
        return get_arr_val_32(node_idx, NODE_CHILD_GT_OFFSET) == -1 && cmp == -1;
    }

    inline void set_resid_ptr(const int32_t node_idx, const void* value) const
    {
        set_arr_val_64(node_idx, reinterpret_cast<int64_t>(value), NODE_RESID_OFFSET);
    }

    inline void* get_resid_ptr(const int32_t node_idx)
    {
        return reinterpret_cast<void*>(get_arr_val_64(node_idx, NODE_RESID_OFFSET));
    }

    inline void set_values_ptr(const int32_t node_idx, const void* value) const
    {
        set_arr_val_64(node_idx, reinterpret_cast<int64_t>(value), NODE_VALUES_OFFSET);
    }

    inline void* get_values_ptr(const int32_t node_idx)
    {
        return reinterpret_cast<void*>(get_arr_val_64(node_idx, NODE_VALUES_OFFSET));
    }

    inline void set_ew_bit(const int32_t node_idx) const
    {
        metadata_arr_[node_idx * METADATA_BYTE_CNT + METADATA_OFFSET] |= 1UL << METADATA_EW_BIT;
    }

    inline void clear_ew_bit(const int32_t node_idx) const
    {
        metadata_arr_[node_idx * METADATA_BYTE_CNT + METADATA_OFFSET] &= ~(1UL << METADATA_EW_BIT);
    }

    inline bool get_ew_bit(const int32_t node_idx) const
    {
        return metadata_arr_[node_idx * METADATA_BYTE_CNT + METADATA_OFFSET] >> METADATA_EW_BIT & 1U;
    }

    inline void init_metadata(const int32_t node_idx) const
    {
        metadata_arr_[node_idx * METADATA_BYTE_CNT + METADATA_OFFSET] = 0x0;
    }

    void node_read_write_test()
    {
        // node 0
        set_arr_val_32(0, 1, NODE_KEY_OFFSET);
        print_arr_bytes();
        auto val = get_arr_val_32(0, NODE_KEY_OFFSET);

        set_arr_val_32(0, 2, NODE_CHILD_LT_OFFSET);
        print_arr_bytes();
        auto lt = get_arr_val_32(0, NODE_CHILD_LT_OFFSET);

        set_arr_val_32(0, 3, NODE_CHILD_EQ_OFFSET);
        print_arr_bytes();
        auto eq = get_arr_val_32(0, NODE_CHILD_EQ_OFFSET);

        set_arr_val_32(0, 4, NODE_CHILD_GT_OFFSET);
        print_arr_bytes();
        auto gt = get_arr_val_32(0, NODE_CHILD_GT_OFFSET);

        set_ew_bit(0);
        print_arr_bytes();
        auto ew = get_ew_bit(0);

        val = get_arr_val_32(0, NODE_KEY_OFFSET);
        lt = get_arr_val_32(0, NODE_CHILD_LT_OFFSET);
        eq = get_arr_val_32(0, NODE_CHILD_EQ_OFFSET);
        gt = get_arr_val_32(0, NODE_CHILD_GT_OFFSET);

        // node 1
        last_tst_idx_++;

        set_arr_val_32(1, 1, NODE_KEY_OFFSET);
        print_arr_bytes();
        val = get_arr_val_32(0, NODE_KEY_OFFSET);

        set_arr_val_32(1, 2, NODE_CHILD_LT_OFFSET);
        print_arr_bytes();
        lt = get_arr_val_32(0, NODE_CHILD_LT_OFFSET);

        set_arr_val_32(1, 3, NODE_CHILD_EQ_OFFSET);
        print_arr_bytes();
        eq = get_arr_val_32(0, NODE_CHILD_EQ_OFFSET);

        set_arr_val_32(1, 4, NODE_CHILD_GT_OFFSET);
        print_arr_bytes();
        gt = get_arr_val_32(0, NODE_CHILD_GT_OFFSET);

        set_ew_bit(1);
        print_arr_bytes();
        ew = get_ew_bit(1);

        val = get_arr_val_32(1, NODE_KEY_OFFSET);
        lt = get_arr_val_32(1, NODE_CHILD_LT_OFFSET);
        eq = get_arr_val_32(1, NODE_CHILD_EQ_OFFSET);
        gt = get_arr_val_32(1, NODE_CHILD_GT_OFFSET);
    }

    void print_node(const int32_t node_idx) const
    {
        std::cout << "\n-------------------------------------\n";
        std::cout << "Node at index: " << node_idx << "\n";
        std::cout << "-------------------------------------\n";
        std::cout << "Value:  " << get_key(node_idx) << "\n";
        std::cout << "LT idx: " << get_lt_idx(node_idx) << "\n";
        std::cout << "EQ idx: " << get_eq_idx(node_idx) << "\n";
        std::cout << "GT idx: " << get_gt_idx(node_idx) << "\n";
        std::cout << "EW bit: " << get_ew_bit(node_idx) << "\n";
        std::cout << "-------------------------------------";
    }

    void print_arr() const
    {
        std::cout << "\n-------------------------------------\n";
        std::cout << "ARRAY" << "\n";
        std::cout << "-------------------------------------\n";

        for (int i = 0; i <= last_tst_idx_; i++)
        {
            std::cout << "Idx: " << i << "\n";
            std::cout << "Val: " << get_key(i) << "\n";
            std::cout << " LT: " << get_lt_idx(i) << "\n";
            std::cout << " EQ: " << get_eq_idx(i) << "\n";
            std::cout << " GT: " << get_gt_idx(i) << "\n";
            std::cout << " EW: " << get_ew_bit(i) << "\n";
        }

        std::cout << "-------------------------------------";
    }

    void print_arr_bytes() const
    {
        std::cout << "\n-------------------------------------\n";
        std::cout << "ARRAY AS BYTES" << "\n";
        std::cout << "-------------------------------------\n";

        for (int i = 0; i <= last_tst_idx_; i++)
        {
            std::cout << "Idx: " << i << "\n";
            for (int b = 0; b < NODE_BYTE_CNT; b++)
            {
                std::cout << ";" << static_cast<int16_t>(tst_arr_[i * NODE_BYTE_CNT + b]);
            }
            std::cout << "\n";
        }

        std::cout << "-------------------------------------";
    }
};

#endif
