#ifndef __TST_FOREST_H__
#define __TST_FOREST_H__

#include <iostream>
#include <map>
#include <vector>
#include "tst.h"

class tst_forest
{
private:
    int32_t _node_stop_cration_depth;
    std::unordered_map<int32_t, tst> tst_forest_;
    std::unordered_map<int64_t, tst> tst_forest_64_;
    int32_t _total_nodes_count;

    int32_t assemble_key(std::vector<int32_t>& vector)
    {
        // all short vectors are stored in TST under the key = 0
        if (vector.size() <= 2)
        {
            return 0;
        }

        int32_t ret_val = vector[0];
        ret_val += vector[1];

        return ret_val;
    }

    inline int32_t assemble_composite_key_32(int32_t k0, int32_t k1)
    {
        if (k0 > 65535 || k1 > 65535)
        {
            return 0;
        }

        int32_t ret_val = k0 << 8;
        ret_val += k1;

        return ret_val;
    }

    int64_t assemble_composite_key_48(std::vector<int32_t>& vector)
    {
        // all short vectors are stored in TST under the key = 0
        if (vector.size() <= 4)
        {
            return 0;
        }

        // all vectors which has key values too large are also stored under key 0
        if (vector[0] > 65535 || 
            vector[1] > 65535 || 
            vector[2] > 65535)
        {
            return 0;
        }

        int32_t ret_val = vector[0] << 16;
        ret_val += vector[1] << 8;
        ret_val += vector[2];

        return ret_val;
    }

    int64_t assemble_composite_key_64(std::vector<int32_t>& vector)
    {
        // all short vectors are stored in TST under the key = 0
        if (vector.size() <= 5)
        {
            return 0;
        }

        // all vectors which has key values too large are also stored under key 0
        if (vector[0] > 65535 || 
            vector[1] > 65535 || 
            vector[2] > 65535 || 
            vector[3] > 65535)
        {
            return 0;
        }

        int32_t ret_val = vector[0] << 24;
        ret_val += vector[1] << 16;
        ret_val += vector[2] << 8;
        ret_val += vector[3];

        return ret_val;
    }

public:
    /// <summary>
    /// Constructor
    /// </summary>
    tst_forest(int32_t node_stop_creation_depth)
    {
        _node_stop_cration_depth = node_stop_creation_depth;
    }

    ~tst_forest()
    {
        for (auto& kvp : tst_forest_)
        {
            kvp.second.clear();
        }

        for (auto& kvp : tst_forest_64_)
        {
            kvp.second.clear();
        }
    }

    int32_t get_tst_forest_size()
    {
        int32_t total_size = 0;
        
        // size of TSTs in the forest
        for (auto& kvp : tst_forest_)
        {
            total_size += kvp.second.get_tst_size();
        }

        // plus size of hash table
        auto forest_size = tst_forest_.size();
        auto val_type_size = sizeof(std::unordered_map<int32_t, tst>::value_type);
        auto hash_table_size = (forest_size * val_type_size) / 1024;
        total_size += hash_table_size;

        return total_size;
    }

    int32_t get_nodes_count()
    {
        return _total_nodes_count;
    }

    void insert(std::vector<std::vector<int32_t>>* keys, std::vector<int32_t>* values)
    {
        auto start_1 = std::chrono::high_resolution_clock::now();

        std::map<int32_t, std::vector<std::vector<int32_t>>> ins_key_map;
        std::map<int32_t, std::vector<int32_t>> ins_val_map;
        std::map<int32_t, int32_t> ins_letters_cnt_map;

        int64_t size_tot = 0;

        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];
            uint32_t map_key = assemble_key(key);

            std::vector<int32_t> new_vect(key.size());
            for (int j = 0; j < key.size(); j++)
            {
                new_vect[j] = key[j];
            }

            ins_letters_cnt_map[map_key] += new_vect.size();
            size_tot += new_vect.size();

            ins_key_map[map_key].push_back(new_vect);
            ins_val_map[map_key].push_back((*values)[i]);
        }

        const auto end_1 = std::chrono::high_resolution_clock::now();
        auto diff_1 = std::chrono::duration_cast<std::chrono::milliseconds>(end_1 - start_1);
        auto diff_t_1 = diff_1.count();

        auto start_2 = std::chrono::high_resolution_clock::now();

        for (auto& kvp : ins_key_map)
        {
            tst tree(ins_letters_cnt_map[kvp.first], _node_stop_cration_depth);
            tree.insert(&kvp.second, &ins_val_map[kvp.first], kvp.second.size());

            tst_forest_.emplace(kvp.first, tree);
        }

        _total_nodes_count = 0;
        int32_t trees_count = 0;
        for (auto& kvp : tst_forest_)
        {
            kvp.second.clean_up_memory();
            _total_nodes_count += kvp.second.get_last_node_idx();
            trees_count++;
        }

        const auto end_2 = std::chrono::high_resolution_clock::now();
        auto diff_2 = std::chrono::duration_cast<std::chrono::milliseconds>(end_2 - start_2);
        auto diff_t_2 = diff_2.count();
    }

    void insert_composite_32(std::vector<std::vector<int32_t>>* keys, std::vector<int32_t>* values)
    {
        auto start_1 = std::chrono::high_resolution_clock::now();

        std::map<int32_t, std::vector<std::vector<int32_t>>> ins_key_map;
        std::map<int32_t, std::vector<int32_t>> ins_val_map;
        std::map<int32_t, int32_t> ins_letters_cnt_map;

        int64_t size_tot = 0;

        for (auto i = 0; i < keys->size(); i++)
        {
            auto key = (*keys)[i];

            uint32_t map_key = 0;
            if (key.size() >= 2)
            {
                map_key = assemble_composite_key_32(key[0], key[1]);
            }

            if (map_key == 0)
            {
                std::vector<int32_t> new_vect(key.size());
                for (int j = 0; j < key.size(); j++)
                {
                    new_vect[j] = key[j];
                }
                ins_letters_cnt_map[map_key] += new_vect.size();
                size_tot += new_vect.size();

                ins_key_map[map_key].push_back(new_vect);
                ins_val_map[map_key].push_back((*values)[i]);
            }
            else
            {
                std::vector<int32_t> new_vect(key.size() - 2);
                for (int j = 0; j < key.size() - 2; j++)
                {
                    new_vect[j] = key[j + 2];
                }
                ins_letters_cnt_map[map_key] += new_vect.size();
                size_tot += new_vect.size();

                ins_key_map[map_key].push_back(new_vect);
                ins_val_map[map_key].push_back((*values)[i]);
            }
        }

        const auto end_1 = std::chrono::high_resolution_clock::now();
        auto diff_1 = std::chrono::duration_cast<std::chrono::milliseconds>(end_1 - start_1);
        auto diff_t_1 = diff_1.count();

        auto start_2 = std::chrono::high_resolution_clock::now();

        int b = 0;
        for (auto& kvp : ins_key_map)
        {
            tst tree(ins_letters_cnt_map[kvp.first], _node_stop_cration_depth);
            tree.insert(&kvp.second, &ins_val_map[kvp.first], kvp.second.size());

            tst_forest_.emplace(kvp.first, tree);

            b++;
            if (b == 1)
            {
                b = 1;
            }
        }

        _total_nodes_count = 0;
        int32_t trees_count = 0;
        for (auto& kvp : tst_forest_)
        {
            kvp.second.clean_up_memory();
            _total_nodes_count += kvp.second.get_last_node_idx();
            trees_count++;
        }

        const auto end_2 = std::chrono::high_resolution_clock::now();
        auto diff_2 = std::chrono::duration_cast<std::chrono::milliseconds>(end_2 - start_2);
        auto diff_t_2 = diff_2.count();
    }

    void insert_composite_48(std::vector<std::vector<int32_t>>* keys, std::vector<int32_t>* values)
    {
        auto start_1 = std::chrono::high_resolution_clock::now();

        std::map<int64_t, std::vector<std::vector<int32_t>>> ins_key_map;
        std::map<int64_t, std::vector<int32_t>> ins_val_map;
        std::map<int64_t, int32_t> ins_letters_cnt_map;

        int64_t size_tot = 0;

        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];
            int64_t map_key = assemble_composite_key_48(key);

            if (map_key == 0)
            {
                std::vector<int32_t> new_vect(key.size());
                for (int j = 0; j < key.size(); j++)
                {
                    new_vect[j] = key[j];
                }
                ins_letters_cnt_map[map_key] += new_vect.size();
                size_tot += new_vect.size();

                ins_key_map[map_key].push_back(new_vect);
                ins_val_map[map_key].push_back((*values)[i]);
            }
            else
            {
                std::vector<int32_t> new_vect(key.size() - 3);
                for (int j = 0; j < key.size() - 3; j++)
                {
                    new_vect[j] = key[j + 3];
                }
                ins_letters_cnt_map[map_key] += new_vect.size();
                size_tot += new_vect.size();

                ins_key_map[map_key].push_back(new_vect);
                ins_val_map[map_key].push_back((*values)[i]);
            }
        }

        const auto end_1 = std::chrono::high_resolution_clock::now();
        auto diff_1 = std::chrono::duration_cast<std::chrono::milliseconds>(end_1 - start_1);
        auto diff_t_1 = diff_1.count();

        auto start_2 = std::chrono::high_resolution_clock::now();

        for (auto& kvp : ins_key_map)
        {
            tst tree(ins_letters_cnt_map[kvp.first], _node_stop_cration_depth);
            tree.insert(&kvp.second, &ins_val_map[kvp.first], kvp.second.size());

            tst_forest_64_.emplace(kvp.first, tree);
        }

        _total_nodes_count = 0;
        int32_t trees_count = 0;
        for (auto& kvp : tst_forest_64_)
        {
            kvp.second.clean_up_memory();
            _total_nodes_count += kvp.second.get_last_node_idx();
            trees_count++;
        }

        const auto end_2 = std::chrono::high_resolution_clock::now();
        auto diff_2 = std::chrono::duration_cast<std::chrono::milliseconds>(end_2 - start_2);
        auto diff_t_2 = diff_2.count();
    }

    void insert_composite_64(std::vector<std::vector<int32_t>>* keys, std::vector<int32_t>* values)
    {
        auto start_1 = std::chrono::high_resolution_clock::now();
        
        std::map<int64_t, std::vector<std::vector<int32_t>>> ins_key_map;
        std::map<int64_t, std::vector<int32_t>> ins_val_map;
        std::map<int64_t, int32_t> ins_letters_cnt_map;

        int64_t size_tot = 0;

        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];
            int64_t map_key = assemble_composite_key_64(key);

            if (map_key == 0)
            {
                std::vector<int32_t> new_vect(key.size());
                for (int j = 0; j < key.size(); j++)
                {
                    new_vect[j] = key[j];
                }
                ins_letters_cnt_map[map_key] += new_vect.size();
                size_tot += new_vect.size();

                ins_key_map[map_key].push_back(new_vect);
                ins_val_map[map_key].push_back((*values)[i]);
            }
            else
            {
                std::vector<int32_t> new_vect(key.size() - 4);
                for (int j = 0; j < key.size() - 4; j++)
                {
                    new_vect[j] = key[j + 4];
                }
                ins_letters_cnt_map[map_key] += new_vect.size();
                size_tot += new_vect.size();

                ins_key_map[map_key].push_back(new_vect);
                ins_val_map[map_key].push_back((*values)[i]);
            }
        }

        const auto end_1 = std::chrono::high_resolution_clock::now();
        auto diff_1 = std::chrono::duration_cast<std::chrono::milliseconds>(end_1 - start_1);
        auto diff_t_1 = diff_1.count();

        auto start_2 = std::chrono::high_resolution_clock::now();

        for (auto& kvp : ins_key_map)
        {
            tst tree(ins_letters_cnt_map[kvp.first], _node_stop_cration_depth);
            tree.insert(&kvp.second, &ins_val_map[kvp.first], kvp.second.size());

            tst_forest_64_.emplace(kvp.first, tree);
        }

        _total_nodes_count = 0;
        int32_t trees_count = 0;
        for (auto& kvp : tst_forest_64_)
        {
            kvp.second.clean_up_memory();
            _total_nodes_count += kvp.second.get_last_node_idx();
            trees_count++;
        }

        const auto end_2 = std::chrono::high_resolution_clock::now();
        auto diff_2 = std::chrono::duration_cast<std::chrono::milliseconds>(end_2 - start_2);
        auto diff_t_2 = diff_2.count();
    }

    std::vector<bool> search(std::vector<std::vector<int32_t>>* keys, int32_t words_size, int64_t &preparation_time, int64_t &search_time)
    {
        auto start_prep = std::chrono::high_resolution_clock::now();

        std::vector<bool> ret_val(words_size);

        std::map<int32_t, std::vector<std::vector<int32_t>>> search_key_map;
        std::map<int32_t, std::vector<int32_t>> idx_map;
        std::map<int32_t, int32_t> sizes_map;
        std::map<int32_t, std::vector<std::vector<int32_t>>> results_map;
        
        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];
            uint32_t map_key = assemble_key(key);

            search_key_map[map_key].push_back(key);
            idx_map[map_key].push_back(i);
        }

        // measure sizes and prepare results map
        for (auto& kvp : search_key_map)
        {
            sizes_map[kvp.first] = kvp.second.size();

            results_map[kvp.first] = std::vector<std::vector<int32_t>>(sizes_map[kvp.first]);
            for (auto& ret_val_iter : results_map[kvp.first])
            {
                ret_val_iter.reserve(10);
            }
        }

        const auto end_prep = std::chrono::high_resolution_clock::now();
        preparation_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_prep - start_prep).count();

        auto start_search = std::chrono::high_resolution_clock::now();

        for (auto& kvp : search_key_map)
        {
            if (tst_forest_.find(kvp.first) == tst_forest_.end())
            {
                for (int i = 0; i < idx_map[kvp.first].size(); i++)
                {
                    ret_val[idx_map[kvp.first][i]] = false;
                }

                continue;
            }

            tst_forest_.at(kvp.first).search(&kvp.second, sizes_map[kvp.first], results_map[kvp.first]);
        }

        const auto end_search = std::chrono::high_resolution_clock::now();
        search_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_search - start_search).count();

        for (auto& kvp : search_key_map)
        {
            for (int i = 0; i < results_map[kvp.first].size(); i++)
            {
                ret_val[idx_map[kvp.first][i]] = results_map[kvp.first][i].size() > 0 ? true : false;
            }
        }

        return ret_val;
    }

    std::vector<bool> search_composite_32(std::vector<std::vector<int32_t>>* keys, int32_t words_size, int64_t &preparation_time, int64_t &search_time)
    {
        auto start_prep = std::chrono::high_resolution_clock::now();

        std::vector<bool> ret_val(words_size);

        std::map<int32_t, std::vector<std::vector<int32_t>>> search_key_map;
        std::map<int32_t, std::vector<int32_t>> idx_map;
        std::map<int32_t, int32_t> sizes_map;
        std::map<int32_t, std::vector<std::vector<int32_t>>> results_map;

        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];

            uint32_t map_key = 0;
            if (key.size() >= 2)
            {
                map_key = assemble_composite_key_32(key[0], key[1]);
            }

            if (map_key == 0)
            {
                search_key_map[map_key].push_back(key);
                idx_map[map_key].push_back(i);
            }
            else
            {
                search_key_map[map_key].push_back(std::vector<int32_t>(key.begin() + 2, key.end()));
                idx_map[map_key].push_back(i);
            }
        }

        // measure sizes and prepare results map
        for (auto& kvp : search_key_map)
        {
            sizes_map[kvp.first] = kvp.second.size();

            results_map[kvp.first] = std::vector<std::vector<int32_t>>(sizes_map[kvp.first]);
            for (auto& ret_val_iter : results_map[kvp.first])
            {
                ret_val_iter.reserve(10);
            }
        }

        const auto end_prep = std::chrono::high_resolution_clock::now();
        preparation_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_prep - start_prep).count();

        auto start_search = std::chrono::high_resolution_clock::now();

        for (auto& kvp : search_key_map)
        {
            if (tst_forest_.find(kvp.first) == tst_forest_.end())
            {
                for (int i = 0; i < idx_map[kvp.first].size(); i++)
                {
                    ret_val[idx_map[kvp.first][i]] = false;
                }

                continue;
            }

            tst_forest_.at(kvp.first).search(&kvp.second, sizes_map[kvp.first], results_map[kvp.first]);

            for (int i = 0; i < results_map[kvp.first].size(); i++)
            {
                ret_val[idx_map[kvp.first][i]] = results_map[kvp.first][i].size() > 0 ? true : false;
            }
        }

        const auto end_search = std::chrono::high_resolution_clock::now();
        search_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_search - start_search).count();

        for (auto& kvp : search_key_map)
        {
            for (int i = 0; i < results_map[kvp.first].size(); i++)
            {
                ret_val[idx_map[kvp.first][i]] = results_map[kvp.first][i].size() > 0 ? true : false;
            }
        }

        return ret_val;
    }

    std::vector<bool> search_composite_48(std::vector<std::vector<int32_t>>* keys, int32_t words_size, int64_t &preparation_time, int64_t &search_time)
    {
        auto start_prep = std::chrono::high_resolution_clock::now();

        std::vector<bool> ret_val(words_size);

        std::map<int64_t, std::vector<std::vector<int32_t>>> search_key_map;
        std::map<int64_t, std::vector<int32_t>> idx_map;
        std::map<int64_t, int32_t> sizes_map;
        std::map<int64_t, std::vector<std::vector<int32_t>>> results_map;

        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];
            int64_t map_key = assemble_composite_key_48(key);

            if (map_key == 0)
            {
                search_key_map[map_key].push_back(key);
                idx_map[map_key].push_back(i);
            }
            else
            {
                search_key_map[map_key].push_back(std::vector<int32_t>(key.begin() + 3, key.end()));
                idx_map[map_key].push_back(i);
            }
        }

        // measure sizes and prepare results map
        for (auto& kvp : search_key_map)
        {
            sizes_map[kvp.first] = kvp.second.size();

            results_map[kvp.first] = std::vector<std::vector<int32_t>>(sizes_map[kvp.first]);
            for (auto& ret_val_iter : results_map[kvp.first])
            {
                ret_val_iter.reserve(10);
            }
        }

        const auto end_prep = std::chrono::high_resolution_clock::now();
        preparation_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_prep - start_prep).count();

        auto start_search = std::chrono::high_resolution_clock::now();

        for (auto& kvp : search_key_map)
        {
            if (tst_forest_64_.find(kvp.first) == tst_forest_64_.end())
            {
                for (int i = 0; i < idx_map[kvp.first].size(); i++)
                {
                    ret_val[idx_map[kvp.first][i]] = false;
                }

                continue;
            }

            tst_forest_.at(kvp.first).search(&kvp.second, sizes_map[kvp.first], results_map[kvp.first]);
        }

        const auto end_search = std::chrono::high_resolution_clock::now();
        search_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_search - start_search).count();

        for (auto& kvp : search_key_map)
        {
            for (int i = 0; i < results_map[kvp.first].size(); i++)
            {
                ret_val[idx_map[kvp.first][i]] = results_map[kvp.first][i].size() > 0 ? true : false;
            }
        }

        return ret_val;
    }

    std::vector<bool> search_composite_64(std::vector<std::vector<int32_t>>* keys, int32_t words_size, int64_t &preparation_time, int64_t &search_time)
    {
        auto start_prep = std::chrono::high_resolution_clock::now();

        std::vector<bool> ret_val(words_size);

        std::map<int64_t, std::vector<std::vector<int32_t>>> search_key_map;
        std::map<int64_t, std::vector<int32_t>> idx_map;
        std::map<int64_t, int32_t> sizes_map;
        std::map<int64_t, std::vector<std::vector<int32_t>>> results_map;

        for (int i = 0; i < keys->size(); i++)
        {
            std::vector<int32_t> key = (*keys)[i];
            int64_t map_key = assemble_composite_key_64(key);

            if (map_key == 0)
            {
                search_key_map[map_key].push_back(key);
                idx_map[map_key].push_back(i);
            }
            else
            {
                search_key_map[map_key].push_back(std::vector<int32_t>(key.begin() + 2, key.end()));
                idx_map[map_key].push_back(i);
            }
        }

        // measure sizes and prepare results map
        for (auto& kvp : search_key_map)
        {
            sizes_map[kvp.first] = kvp.second.size();

            results_map[kvp.first] = std::vector<std::vector<int32_t>>(sizes_map[kvp.first]);
            for (auto& ret_val_iter : results_map[kvp.first])
            {
                ret_val_iter.reserve(10);
            }
        }

        const auto end_prep = std::chrono::high_resolution_clock::now();
        preparation_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_prep - start_prep).count();

        auto start_search = std::chrono::high_resolution_clock::now();

        for (auto& kvp : search_key_map)
        {
            if (tst_forest_64_.find(kvp.first) == tst_forest_64_.end())
            {
                for (int i = 0; i < idx_map[kvp.first].size(); i++)
                {
                    ret_val[idx_map[kvp.first][i]] = false;
                }

                continue;
            }

            tst_forest_.at(kvp.first).search(&kvp.second, sizes_map[kvp.first], results_map[kvp.first]);
        }

        const auto end_search = std::chrono::high_resolution_clock::now();
        search_time = std::chrono::duration_cast<std::chrono::milliseconds>(end_search - start_search).count();

        for (auto& kvp : search_key_map)
        {
            for (int i = 0; i < results_map[kvp.first].size(); i++)
            {
                ret_val[idx_map[kvp.first][i]] = results_map[kvp.first][i].size() > 0 ? true : false;
            }
        }

        return ret_val;
    }
};

#endif
