#ifndef LENGTHENCODER_H
#define LENGTHENCODER_H

#include <type_traits>
#include <cmath>
#include <cstdint>

#include "../Configuration/configuration.h"

template<typename TIN, typename TOUT>
TOUT LinearEncode(TIN value, float epsilon)
{
  static_assert(std::is_arithmetic<TIN>::value, "TIN must be numeric type");
  static_assert(std::is_arithmetic<TOUT>::value, "TOUT must be numeric type");

  if (epsilon == 0)
  {
    epsilon = 1;
  }

  return round(value / epsilon);
}

template<typename TIN, typename TOUT>
TOUT LogEncode(TIN value, int16_t base, float epsilon)
{
  static_assert(std::is_arithmetic<TIN>::value, "TIN must be numeric type");
  static_assert(std::is_arithmetic<TOUT>::value, "TOUT must be numeric type");

  if (value == 0 || base == 0)
  {
    return 0;
  }

  if (epsilon == 0)
  {
    epsilon = 1;
  }

  auto logCoeff = static_cast<float>(log(value) / log(base));

  if (logCoeff == 0)
  {
    logCoeff = 1;
  }

  epsilon = epsilon * logCoeff;

  return LinearEncode<TIN, TOUT>(value, epsilon);
}

template<typename TIN, typename TOUT>
TOUT PowEncode(TIN value, float pow, float epsilon)
{
  static_assert(std::is_arithmetic<TIN>::value, "TIN must be numeric type");
  static_assert(std::is_arithmetic<TOUT>::value, "TOUT must be numeric type");

  if (value == 0)
  {
    return 0; 
  }

  if (epsilon == 0)
  {
    epsilon = 1;
  }

  auto powCoeff = static_cast<float>(std::pow(value, pow));

  if (powCoeff == 0)
  {
    powCoeff = 1;
  }

  epsilon = epsilon * powCoeff;

  return LinearEncode<TIN, TOUT>(value, epsilon);
}

template<typename TIN, typename TOUT>
TOUT PolEncode(TIN value, float maxValue, float a, float b, float c, float d, float e, float epsilon)
{
  static_assert(std::is_arithmetic<TIN>::value, "TIN must be numeric type");
  static_assert(std::is_arithmetic<TOUT>::value, "TOUT must be numeric type");

  if (value == 0)
  {
    return 0; 
  }

  if (epsilon == 0)
  {
    epsilon = 1;
  }

  auto valRatio = (float)value/(float)maxValue;
  auto coeff = a * std::pow(valRatio, 5) + b * std::pow(valRatio, 4) + c * std::pow(valRatio, 3) + d * std::pow(valRatio, 2) + e * valRatio;

  if (coeff == 0)
  {
    coeff = 1;
  }

  epsilon = epsilon * coeff;

  return LinearEncode<TIN, TOUT>(value, epsilon);
}

template<typename T>
T EncodeValue(std::string encoder_method, T value, T maxValueForPol)
{
	if (encoder_method == "LIN")
	{
		return LinearEncode<T, T>(value, _LengthsEncoderConf_Epsilon);
	}
	if (encoder_method == "LOG")
	{
		return LogEncode<T, T>(value, _LengthsEncoderConf_LogBase, _LengthsEncoderConf_Epsilon);
	}
	if (encoder_method == "POW")
	{
		return PowEncode<T, T>(value, _LengthsEncoderConf_Power, _LengthsEncoderConf_Epsilon);
	}
	if (encoder_method == "POL")
	{
		return PolEncode<T, T>(value, maxValueForPol, _LengthsEncoderConf_PolCoeffA, _LengthsEncoderConf_PolCoeffB, _LengthsEncoderConf_PolCoeffC,
			_LengthsEncoderConf_PolCoeffD, _LengthsEncoderConf_PolCoeffE, _LengthsEncoderConf_Epsilon);
	}

	return 0;
}

template<typename T>
std::vector<T> EncodeVector(std::string encoder_method, std::vector<T> vector, T maxValueForPol, T& valMin, T& valMax, T& encValMin, T& encValMax, T& valCnt)
{
  std::vector<T> ret_vect;
  ret_vect.reserve(vector.size());

	for (int64_t i = 0; i < vector.size(); i++)
	{
		if (valMin > vector[i])
		{
			valMin = vector[i];
		}

		if (valMax < vector[i])
		{
			valMax = vector[i];
		}

    auto encoded_value = EncodeValue<T>(encoder_method, vector[i], maxValueForPol);
		ret_vect.push_back(encoded_value);

		if (encValMin > vector[i])
		{
			encValMin = vector[i];
		}

		if (encValMax < vector[i])
		{
			encValMax = vector[i];
		}

		valCnt++;
	}

  return ret_vect;
}

#endif