#ifndef __TEST_UTILS_H__
#define __TEST_UTILS_H__

#include <string>
#include <iostream>
#include "lengthencoder.h"
#include "Configuration/configuration.h"

using DT = int32_t;

bool is_test_enabled(std::string test)
{
	for (auto& test_enable : _TestEnable)
	{
		if (test_enable == test)
		{
			return true;
		}
	}

	return false;
}

long get_process_size()
{
	long s = -1;

   	FILE *f = fopen("/proc/self/statm", "r");
   
	if (!f) 
	{
		return s;
	}

	fscanf(f, "%ld", &s);
   	fclose(f);	

   	return s * _System_PageSize;
}

template<typename T>
void calculate_label_distances_and_offsets(std::vector<double> input_data, std::vector<T>& distances, std::vector<T>& offsets, T& max_label_distance)
{
   	for (auto i = 0; i < input_data.size() - 1; i++)
	{
		auto dist = static_cast<DT>(input_data[i + 1] - input_data[i]);
		
		if (dist < 0)
		{
			continue;
		}
		
		distances.push_back(dist);
		offsets.push_back(input_data[i + 1]);
		max_label_distance = max_label_distance < dist ? dist : max_label_distance;
	}
}

template<typename T>
void calculate_label_distances_and_offsets(std::vector<T> input_data, std::vector<T>& distances, std::vector<T>& offsets, T& max_label_distance)
{
   	for (auto i = 0; i < input_data.size() - 1; i++)
	{
		auto dist = static_cast<DT>(input_data[i + 1] - input_data[i]);
		distances.push_back(dist);
		offsets.push_back(input_data[i + 1]);
		max_label_distance = max_label_distance < dist ? dist : max_label_distance;
	}
}

template<typename T>
std::vector<T> encode_vector(std::string encoder_method, std::vector<T> values, T max_label_distance)
{
   	auto val_min = (std::numeric_limits<T>::max)();
   	auto val_max = 0;
	auto enc_val_min = (std::numeric_limits<T>::max)();
	auto enc_val_max = 0;
	auto val_cnt = 0;

	auto encoded_values = EncodeVector<T>(encoder_method, values, max_label_distance, val_min, val_max, enc_val_min, enc_val_max, val_cnt);

   	return encoded_values;
}

template<typename T>
void create_k_mers(int32_t k_mer_length, std::vector<T> values, std::vector<T> offsets, std::vector<std::vector<T>>& k_mers, std::vector<T>& k_mer_offsets)
{
	k_mers.clear();
	k_mers.reserve(values.size());
	
	if (offsets.size() > 0)
	{
		k_mer_offsets.clear();
		k_mer_offsets.reserve(values.size());
	}

	for (auto i = 0; i < values.size() - k_mer_length + 1; i++)
	{
		// create k-mer
		std::vector<DT> k_mer;
		k_mer.reserve(k_mer_length);
		for (auto j = i; j < i + k_mer_length; j++)
		{
			k_mer.push_back(values[j]);
		}

		// do not create k-mers shorter than shortest molecules from BNX
		if (k_mer.size() < _GeneralSettings_BnxMoleculeMinLabelsCount)
		{
			continue;
		}

		k_mers.push_back(k_mer);

		if (offsets.size() > i)
		{
			k_mer_offsets.push_back(offsets[i]);
		}

		// create k-mer substring if needed
		if (_GeneralSettings_KMersSubstrings)
		{
			auto begin = k_mer.begin();
			auto end = k_mer.end();

			while (begin < end - 1)
			{
				std::vector<DT> substring_vect(begin, end);
				k_mers.push_back(substring_vect);
				
				if (offsets.size() > 0)
				{
					k_mer_offsets.push_back(offsets[i]);
				}				
				
				++begin;
			}
		}		
	}
}

std::string get_test_id(test_definition test_def)
{
	return std::to_string(test_def._k_mer_size) + "-" + test_def._encode_method ;
}

std::string create_log_message(std::string test_id, std::string test_type, int64_t indexing_time, int64_t search_time_tot,
	int64_t prep_search_time, int64_t raw_search_time, int64_t found_cnt, int64_t not_found_cnt, int64_t size_kb)
{
	return test_id + ";" + test_type + ";" + std::to_string(raw_search_time) + ";" + std::to_string(search_time_tot) + ";" + 
		std::to_string(prep_search_time) + ";" + std::to_string(indexing_time) + ";" + std::to_string(size_kb);
}

std::vector<int32_t> calc_alphabet_counts(std::vector<int32_t> vect)
{
	int32_t max = 0;
	for (auto& val : vect)
	{
		if (max < val)
		{
			max = val;
		}
	}

	std::vector<int32_t> alphabet_counts(max + 1);
	for (int32_t i = 0; i < alphabet_counts.size(); i++)
	{
		alphabet_counts[i] = 0;
	}

	for (auto& val : vect)
	{
		alphabet_counts[val]++;
	}

	return alphabet_counts;
}

std::vector<int32_t> calc_alphabet_counts_without_max(std::vector<int32_t> vect, std::ofstream& log_file)
{
	int32_t max1 = 0;
	for (auto& val : vect)
	{
		if (max1 < val)
		{
			max1 = val;
		}
	}

	int32_t max = 0;
	for (auto& val : vect)
	{
		if (max < val && val != max1)
		{
			max = val;
		}
	}

	std::vector<int32_t> alphabet_counts(max + 1);
	for (int32_t i = 0; i < alphabet_counts.size(); i++)
	{
		alphabet_counts[i] = 0;
	}

	for (auto& val : vect)
	{
		if (val <= max)
		{
			alphabet_counts[val]++;
		}
		else
		{
			log_file << "Max left out: " << val << endl;
		}
	}

	return alphabet_counts;
}

int32_t get_alphabet_size(std::vector<int32_t> alphabet_counts)
{
	int32_t cnt = 0;
	for (int32_t i = 0; i < alphabet_counts.size(); i++)
	{
		if (alphabet_counts[i] > 0)
		{
			cnt++;
		}
	}

	return cnt;
}

void log_histogram(std::vector<int32_t> alphabet_counts, int32_t intervals, std::ofstream& log_file)
{
	std::vector<int32_t> histogram;

	auto step = (alphabet_counts.size() - 1) / intervals;
	log_file << endl << "Histogram with step = " << step << ": "<< endl << endl;

	int idx = 0;
	for (auto& count : alphabet_counts)
	{
		auto pos = (int32_t)(idx / step);
		
		if (pos >= histogram.size())
		{
			histogram.push_back(0);
		}
		
		histogram[pos] += count;
		idx++;
	}

	int cnt = 0;
	for (auto& hist : histogram)
	{
		log_file << cnt << "," << cnt * step << "," << (cnt + 1) * step << "," << hist << endl;
		cnt++;
	}
}

#endif