///====================================================================================================
///@file      Loader.h
///@author    Petr Gajdoš (petr.gajdos@vsb.cz)
///@brief     This class provides basic methods to load/parse known data structures. See bioStructures.h
///           for more details about data types and aliases. Beyond methods to standard formats (Bionano), 
///           there are methods to load custom user-defined formats.
///@version   1.0
///@date      2020-01-16
///
///@copyright Copyright (c) 2020
///
///====================================================================================================

#pragma once

#include <charconv>
#include <vector>
#include <fstream>
#include <sstream>
#include <iostream>
#include <string_view>

#include <utils/fileUtils.h>
#include <utils/stdVectorUtils.h>
#include <utils/stdTupleUtils.h>
#include <utils/stringUtils.h>
#include <utils/parsing.h>

#include "bioStructures.h"

using namespace pg;

using std::string;
using std::string_view;
using std::vector;
using std::ostream;
using std::cout;
using std::endl;
using std::string_view;
using std::streampos;
using std::ofstream;


class Loader 
{
public: 
	using filterFn = std::function<bool(const char* start, char* end)>;

private:
    template<typename RECORD_TYPE>
    static std::size_t parse(char* buffer, const string_view& delimiters, vector<RECORD_TYPE>& records, vector<string>& header, filterFn rowFilterFunction = nullptr);
    template<typename RECORD_TYPE>
    static std::size_t parseColumns(char* buffer, const string_view& delimiters, const vector<uint32_t>& ascendingIndices, vector<RECORD_TYPE>& records, vector<string>& header, filterFn rowFilterFunction = nullptr);

    template<typename RECORD_TYPE>
    static std::size_t parseBNX(char* buffer, const string_view& delimiters, vector<RECORD_TYPE>& records, vector<string>& header, filterFn rowFilterFunction = nullptr) { return 0; }

public:
    //BIONANO GENOMICS FORMATS
    template<typename RECORD_TYPE>
    static bool loadBionano(const string& fileName, vector<RECORD_TYPE>& records, vector<string>& headers, filterFn rowFilterFunction = nullptr);
    template<typename RECORD_TYPE>
    static bool loadBionanoColumns(const string& fileName, vector<RECORD_TYPE>& records, vector<string>& headers, const vector<uint32_t>* ascendingIndices, filterFn rowFilterFunction = nullptr);

	static bool parseIDSfromBNX(const string& fileName, vector<std::uint32_t>& ids);

    //STANDARDIZED FORMATS
	template<typename RECORD_TYPE>
    static bool parseMarkersFromFasta(const string& fileName, const string& outputPath, const string& markerSequence, vector<RECORD_TYPE>& rawData, vector<SequenceInfo>& sequenceInfo);
	template<typename RECORD_TYPE>
    static bool parseMarkersFromFastaSplitParts(const string& fileName, const string& outputPath, const string& markerSequence, vector<RECORD_TYPE>& rawData, vector<SequenceInfo>& sequenceInfo);
    
    //LOCAL FORMATS
    static bool loadCMAPSequences(const char* fileName, vector<float>& rawData, vector<SequenceInfo>& sequenceInfo);
    static bool loadFinalAlignments(const char* fileName, vector<AlignmentData>& alignmentData);
};

///====================================================================================================
///@brief XMAPPair - specialization for parseValue. See the general declaration in <utils/parsing.h> 
///====================================================================================================
template<> inline void pg::parsing::parseValue(const char* start, char* end, XMAPPair &value)
{
    const char* comma = start;
    for (;comma != end; comma++)
    {
        if (*comma == ',') break;
    }

    //No error checking !!!!	
    std::from_chars<uint32_t>(start, start + (comma-start), value.first);
    std::from_chars<uint32_t>(comma+1, comma + (end-comma), value.second);
    // value.first = static_cast<std::uint32_t>(atol(start));
    // value.second = static_cast<std::uint32_t>(atol(comma+1));
}

///====================================================================================================
///@brief Calling method to load bionano files. See bioStructures.h for more details about supported RECORD_TYPES.
///
///@tparam RECORD_TYPE 
///@param fileName File to load.
///@param records Ouput vector of records (tuples)
///@param header Output vector of strings (column names) contained in the file headers.
///@param rowFilterFunction Function that is used to filter out certain rows/records based on value in a given column. Default is nullptr = all rows are accepted.
///@return bool TRUE if loading was ok, otherwise FALSE.
///====================================================================================================
template<typename RECORD_TYPE>
bool Loader::loadBionano(const string& fileName, vector<RECORD_TYPE>& records, vector<string>& header, filterFn rowFilterFunction)
{
    try 
    {
        char* buffer = nullptr;
        std::streampos fileSize;

        if (!pg::fileutils::getBuffer(fileName, buffer, fileSize)) return false;

        const string_view delimiters = "\t";
        
		if constexpr (pg::stdtupleutils::is_tuple<RECORD_TYPE>::value)
		{
			parse(buffer, delimiters, records, header, rowFilterFunction);
		}
		else
		{
			parseBNX(buffer, delimiters, records, header, rowFilterFunction);
		}

        ::operator delete(buffer);
        return true;    
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << std::endl;
    }
    return false;
}

///====================================================================================================
///@brief Calling method to load bionano files. See bioStructures.h for more details about supported RECORD_TYPES.
///		  Only selected columns will be read.
///@tparam RECORD_TYPE 
///@param fileName File to load.
///@param records Ouput vector of records (tuples)
///@param header Output vector of strings (column names) contained in the file headers.
///@param ascendingIndices Vector of ascending indices of columns to be read.
///@param rowFilterFunction Function that is used to filter out certain rows/records based on value in a given column. Default is nullptr = all rows are accepted.
///@return bool TRUE if loading was ok, otherwise FALSE.
///====================================================================================================
template<typename RECORD_TYPE>
bool Loader::loadBionanoColumns(const string& fileName, vector<RECORD_TYPE>& records, vector<string>& header, const vector<uint32_t>* ascendingIndices, filterFn rowFilterFunction)
{
    try 
    {
        char* buffer = nullptr;
        std::streampos fileSize;

        if (!pg::fileutils::getBuffer(fileName, buffer, fileSize)) return false;

        const string_view delimiters = "\t";

		if constexpr (pg::stdtupleutils::is_tuple<RECORD_TYPE>::value)
		{
			parseColumns(buffer, delimiters, *ascendingIndices, records, header, rowFilterFunction);
		}

        ::operator delete(buffer);
        return true;    
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << std::endl;
    }
    return false;
}

///====================================================================================================
///@brief As the bionano data formats are very similar, this method parses all data rows. 
///		  Each row is represented by a single tuple of RECORD_TYPE.
///
///@tparam RECORD_TYPE 
///@param buffer File buffer data.
///@param delimiters Delimiters to split row data into elements.
///@param records Ouput vector of records (tuples)
///@param header Output vector of strings (column names) contained in the file headers.
///@param rowFilterFunction Function that is used to filter out certain rows/records based on value in a given column. Default is nullptr = all rows are accepted.
///@return std::size_t Number of loaded records/rows.
///====================================================================================================
template<typename RECORD_TYPE>
std::size_t Loader::parse(char* buffer, const string_view& delimiters, vector<RECORD_TYPE>& records, vector<string>& header, filterFn rowFilterFunction)
{
	constexpr unsigned C = std::tuple_size_v<RECORD_TYPE>;

	string_view sw(buffer);
	string_view::size_type start = 0;
	string_view::size_type pos = sw.find('\n', start);      //pos = sw.find_first_of('\n', start);

	bool headerFound = false;
	bool filterPassed = true;
	RECORD_TYPE tmpRecord;

    const char *h = "#h";

	while((pos != string_view::npos) && (headerFound==false))
	{
		if (pos <= start) break;

		if (strncmp(buffer + start, h, 2)==0)           //if (*(buffer + start) == '#' && *(buffer + start+1) == 'h')
		{
			header.reserve(C);
			headerFound = pg::stdvectorutils::parseVector(delimiters, buffer + start + 3, buffer+pos, header);		//+3 skips: # + h + separator
		}

		start = pos + 1;
		pos = sw.find('\n', start);                     //pos = sw.find_first_of('\n', start);
	}
	if (!headerFound) return 0;

	while (pos != string_view::npos)
	{
		if (pos <= start) break;

		if (*(buffer + start) != '#')
		{
			if (rowFilterFunction != nullptr)
			{
				filterPassed = rowFilterFunction(buffer + start, buffer + pos);
			}
			if (filterPassed)
			{
				if (!pg::stdtupleutils::parseTuple(delimiters, buffer + start, buffer + pos, tmpRecord))
				{
					cout << "Bad record on line " << records.size() << " in the input file ... " << endl;
				}
				else
				{
					records.emplace_back(tmpRecord);
				}
			}
		}

		start = pos + 1;
		pos = sw.find('\n', start);                          //pos = sw.find_first_of('\n', start);
	}

	records.shrink_to_fit();
	return records.size();
}

///====================================================================================================
///@brief As the bionano data formats are very similar, this method parses data of selected columns of all rows. 
///		  Obtained records are represented by tuples of RECORD_TYPE.
///
///@tparam RECORD_TYPE 
///@param buffer File buffer data.
///@param delimiters Delimiters to split row data into elements.
///@param ascendingIndices Vector of ascending indices of columns to be read.
///@param records Ouput vector of records (tuples)
///@param header Output vector of strings (column names) contained in the file headers.
///@param rowFilterFunction Function that is used to filter out certain rows/records based on value in a given column. Default is nullptr = all rows are accepted.
///@return std::size_t Number of loaded records/rows.
///====================================================================================================
template<typename RECORD_TYPE>
std::size_t Loader::parseColumns(char* buffer, const string_view& delimiters, const vector<uint32_t>& ascendingIndices, vector<RECORD_TYPE>& records, vector<string>& header, filterFn rowFilterFunction)
{
	constexpr unsigned C = std::tuple_size_v<RECORD_TYPE>;

	string_view sw(buffer);
	string_view::size_type start = 0;
	string_view::size_type pos = sw.find('\n', start);      //pos = sw.find_first_of('\n', start);

	bool headerFound = false;
	bool filterPassed = true;
	RECORD_TYPE tmpRecord;

    const char *h = "#h";

	while((pos != string_view::npos) && (headerFound==false))
	{
		if (pos <= start) break;

		if (strncmp(buffer + start, h, 2)==0)           //if (*(buffer + start) == '#' && *(buffer + start+1) == 'h')
		{
			header.reserve(C);
			headerFound = pg::stdvectorutils::parseVector(delimiters, buffer + start + 3, buffer+pos, header);		//+3 skips: # + h + separator
		}

		start = pos + 1;
		pos = sw.find('\n', start);                     //pos = sw.find_first_of('\n', start);
	}
	if (!headerFound) return 0;

	while (pos != string_view::npos)
	{
		if (pos <= start) break;

		if (*(buffer + start) != '#')
		{
			if (rowFilterFunction != nullptr)
			{
				filterPassed = rowFilterFunction(buffer + start, buffer + pos);
			}
			if (filterPassed)
			{
				if (!pg::stdtupleutils::parseSubTuple(delimiters, ascendingIndices, buffer + start, buffer + pos, tmpRecord))
				{
					cout << "Bad record on line " << records.size() << " in the input file ... " << endl;
				}
				else
				{
					records.emplace_back(tmpRecord);
				}
			}
		}

		start = pos + 1;
		pos = sw.find('\n', start);                          //pos = sw.find_first_of('\n', start);
	}

	records.shrink_to_fit();
	return records.size();
}

///====================================================================================================
///@brief BNX if one of the files that are defined in out of bionano scope. Order of data rows is a little bit
//        complicated. That's why parseBNX function is defined above and this is just the specialized function
///       providing loading of a certain BNX file.
///
///@tparam  bnx_v13 for this particular specialization.
///@param buffer File buffer data.
///@param delimiters Delimiters to split row data into elements.
///@param records Ouput vector of records (tuples)
///@param header Output vector of strings (column names) contained in the file headers.
///@param rowFilterFunction Function that is used to filter out certain rows/records based on value in a given column. Default is nullptr = all rows are accepted.
///@return std::size_t Number of loaded records/rows.
///====================================================================================================
template<>
inline std::size_t Loader::parseBNX<bnx_v13>(char* buffer, const string_view& delimiters, vector<bnx_v13>& records, vector<string>& header, filterFn rowFilterFunction)
{
    // static_assert(has_member(RECORD_TYPE, base), "RECORD_TYPE must contain base member");
    // static_assert(has_member(RECORD_TYPE, positions), "RECORD_TYPE must contain positions member");
    // static_assert(has_member(RECORD_TYPE, qx11), "RECORD_TYPE must contain qx11 member");
    // static_assert(has_member(RECORD_TYPE, qx12), "RECORD_TYPE must contain qx12 member");

	string_view sw(buffer);
	string_view::size_type start = 0;
	string_view::size_type pos = sw.find('\n', start);      //pos = sw.find_first_of('\n', start);

	bool headerFound = false;
	bool filterPassed = true;
	bnx_v13 tmpRecord;

    const char* h = "#0h";

	while((pos != string_view::npos) && (headerFound==false))
	{
		if (pos <= start) break;

        if (strncmp(buffer + start, h, 3)==0)               //if (*(buffer + start) == '#' && *(buffer + start+1) == '0' && *(buffer + start+1) == 'h')
		{
			headerFound = pg::stdvectorutils::parseVector(delimiters, buffer + start + 4, buffer+pos, header);		//+3 skips: # + h + separator
		}

		start = pos + 1;
		pos = sw.find('\n', start);                         //pos = sw.find_first_of('\n', start);
	}
	if (!headerFound) return 0;

	while (pos != string_view::npos)
	{
		if (pos <= start) break;

		if (*(buffer + start) == '0')
		{
			if (rowFilterFunction != nullptr)
			{
				filterPassed = rowFilterFunction(buffer + start, buffer + pos);
			}
			if (filterPassed)
			{
				if (!pg::stdtupleutils::parseTuple(delimiters, buffer + start, buffer + pos, tmpRecord.base))
				{
						cout << "Bad record on line " << records.size() << " in the input file ... " << endl;
				}
				else
				{
					if (records.size()>0)
					{
						records.back().qx11.second.shrink_to_fit();
						records.back().qx12.second.shrink_to_fit();
					}
					tmpRecord.positions.second.reserve(std::get<5>(tmpRecord.base)+1);
					tmpRecord.qx11.second.reserve(std::get<5>(tmpRecord.base));
					tmpRecord.qx12.second.reserve(std::get<5>(tmpRecord.base));
					records.emplace_back(tmpRecord);
				}
			}
		} 
        else if ((*(buffer + start) == '1')&&(filterPassed))
        {
            pg::parsing::parseValue(buffer + start, buffer + pos, records.back().positions.first);
            start += 2;
            pg::stdvectorutils::parseVector(delimiters, buffer + start, buffer+pos, records.back().positions.second);
        }
        else if ((strncmp(buffer + start, "QX11", 4)==0)&&(filterPassed))
        {
            pg::parsing::parseValue(delimiters, buffer + start, buffer + pos, records.back().qx11.first);
            start += 5;
            pg::stdvectorutils::parseVector(delimiters, buffer + start, buffer+pos, records.back().qx11.second);
        }
        else if ((strncmp(buffer + start, "QX12", 4)==0)&&(filterPassed))
        {
            pg::parsing::parseValue(delimiters, buffer + start, buffer + pos, records.back().qx12.first);
            start += 5;
            pg::stdvectorutils::parseVector(delimiters, buffer + start, buffer+pos, records.back().qx12.second);
        }
		start = pos + 1;
		pos = sw.find('\n', start);                             //pos = sw.find_first_of('\n', start);
	}

    records.back().qx11.second.shrink_to_fit();
    records.back().qx12.second.shrink_to_fit();
    records.shrink_to_fit();
	return records.size();
}

template<typename RECORD_TYPE>
bool Loader::parseMarkersFromFasta(const string& fileName, const string& outputPath, const string& markerSequence, vector<RECORD_TYPE>& rawData, vector<SequenceInfo>& sequenceInfo)
{
	char* buffer = nullptr;
	streampos fileSize;

	if (!pg::fileutils::getBuffer(fileName, buffer, fileSize))
	{
		return false;
	}
	
	mode_t nMode = 0733; // UNIX style permissions
	pg::fileutils::removePath(outputPath);
	pg::fileutils::createPath(nMode, "", outputPath);

	char *end = buffer + fileSize;
	char *from = nullptr;

	size_t counter = 1;
	string recordName;
	string outputFileName;
	ofstream ofs;

	size_t markerLength = markerSequence.length();
	size_t markerPosition = 0;

	SequenceInfo* si = nullptr;
	uint32_t offset = 0;
	uint32_t fileCounter = 0;
	bool isAlternate = false;

	for (char *ptr = buffer; ptr != end; ptr++)
	{
		if (*ptr == '>') 					// '>' starts every description row in the fasta format
		{
			from = ptr + 1;
			while (*ptr != '\n')
			{
				if (*ptr == '_') isAlternate = true;
				ptr++;
			}

			if (isAlternate)				// Skip the whole alternate molecule
			{
				while ((*ptr != '>')&&(ptr < end))
				{
					ptr++;
				}
				ptr--;
				isAlternate = false;
				continue;
			}

			recordName = string(from, ptr - from);
			cout << recordName << endl;

			outputFileName = outputPath + "/" + recordName + ".txt";
			if (!pg::fileutils::createOFS(ofs, outputFileName)) break;

			sequenceInfo.push_back({fileCounter, offset, 0});
			si = &sequenceInfo.back();
			// si->id = fileCounter;
			// si->offset = offset;
			// si->length = 0;

			markerPosition = 0;
			fileCounter++;
			counter = 0;
			from = ptr + 1;

			continue;
		}

		if (*ptr == '\n')
		{
			if (ofs.is_open())
			{
				ofs.write(from, ptr - from);
			}
			from = ptr + 1;
			continue;
		}

		switch (*ptr)						//convert alphabet to lower cases
		{
		case 'N':
			*ptr = '.';
			break;
		case 'A':
			*ptr = 'a';
			break;
		case 'G':
			*ptr = 'g';
			break;
		case 'T':
			*ptr = 't';
			break;
		case 'C':
			*ptr = 'c';
			break;
		default:
			break;
		}
		++counter;

		if (markerSequence[markerPosition] == *ptr)
		{
			++markerPosition;
		
			if (markerPosition == markerLength)
			{
				si->length++;
				offset++;
				rawData.emplace_back(static_cast<RECORD_TYPE>(counter - markerLength));
				markerPosition = 0;
			}
		}
		else
		{
			markerPosition = (markerSequence[0] == *ptr) ? 1 : 0;
		}
	}
	if (ofs.is_open())
	{
		ofs.flush();
		ofs.close();
	}
	rawData.shrink_to_fit();
	sequenceInfo.shrink_to_fit();

	::operator delete(buffer);
	return true;
}

template<typename RECORD_TYPE>
bool Loader::parseMarkersFromFastaSplitParts(const string& fileName, const string& outputPath, const string& markerSequence, vector<RECORD_TYPE>& rawData, vector<SequenceInfo>& sequenceInfo)
{
	char* buffer = nullptr;
	streampos fileSize;

	if (!pg::fileutils::getBuffer(fileName, buffer, fileSize))
	{
		return false;
	}
	
	mode_t nMode = 0733; // UNIX style permissions
	pg::fileutils::removePath(outputPath);
	pg::fileutils::createPath(nMode, "", outputPath);

	char *end = buffer + fileSize;
	char *from = nullptr;

	size_t counter = 1;
	uint32_t fileCounter = 0;
	string recordName;
	string outputFileName;
	ofstream ofs;

	size_t markerLength = markerSequence.length();
	size_t markerPosition = 0;

	SequenceInfo* si = nullptr;
	uint32_t offset = 0;

	for (char *ptr = buffer; ptr != end; ptr++)
	{
		if (*ptr == '>') 					// '>' starts every description row in the fasta format
		{
			from = ptr;
			while (*ptr != '\n')
			{
				ptr++;
			}

			recordName = string(from + 1, ptr - from - 1);
			cout << recordName << endl;

			if (pg::fileutils::createPath(nMode, outputPath, recordName) != 0)	break;
			counter = 1;
			fileCounter = 0;
			from = nullptr;

			continue;
		}

		if (*ptr == 'N')
		{
			if (from != nullptr) 			//process record data
			{
				if (ofs.is_open())
				{
					ofs.write(from, ptr - from);
				}
				from = nullptr;
			}
			while ((*ptr == 'N') || (*ptr == '\n'))
			{
				if (*ptr == 'N')
					++counter;
				ptr++;
			}
			ptr--;

			continue;
		}

		if (from == nullptr)
		{
			from = ptr;
			outputFileName = outputPath + "/" + recordName + "/part" + std::to_string(fileCounter) + ".txt";
			if (!pg::fileutils::createOFS(ofs, outputFileName)) break;
			ofs << counter << endl;

			sequenceInfo.push_back({fileCounter, offset, 0});
			si = &sequenceInfo.back();
			// si->id = fileCounter;
			// si->offset = offset;
			// si->length = 0;

			markerPosition = 0;
			fileCounter++;
		}

		if (*ptr == '\n')
		{
			if (from != nullptr)
			{
				if (ofs.is_open())
				{
					ofs.write(from, ptr - from);
				}
				from = ptr + 1;
				continue;
			}
		}

		switch (*ptr)						//convert alphabet to lower cases
		{
		case 'A':
			*ptr = 'a';
			break;
		case 'G':
			*ptr = 'g';
			break;
		case 'T':
			*ptr = 't';
			break;
		case 'C':
			*ptr = 'c';
			break;
		default:
			break;
		}
		++counter;

		if (markerSequence[markerPosition] == *ptr)
		{
			++markerPosition;
		
			if (markerPosition == markerLength)
			{
				si->length++;
				offset++;
				rawData.emplace_back(static_cast<RECORD_TYPE>(counter - markerLength));
				markerPosition = 0;
			}
		}
		else
		{
			markerPosition = (markerSequence[0] == *ptr) ? 1 : 0;
		}
	}
	if (ofs.is_open())
	{
		ofs.flush();
		ofs.close();
	}
	rawData.shrink_to_fit();
	sequenceInfo.shrink_to_fit();

	::operator delete(buffer);
	return true;
}