///====================================================================================================
///@file      Sequencer.h
///@author    Petr Gajdoš (petr.gajdos@vsb.cz)
///@brief     Sequencer provides basic methods to work with sequences of T data type, i.e. scaling, denoising, 
///			  reversing, etc.
///@version   1.0
///@date      2020-01-12
///
///@copyright Copyright (c) 2020
///
///====================================================================================================
#pragma once

#include <vector>
#include <limits>

#include <utils/stdVectorUtils.h>
#include <utils/slidingWindowUtils.h>
#include <ProgressBar.h>
#include <Comparator.h>

#include "bioStructures.h"

using std::vector;
using namespace pg;

template<typename T>
class Sequencer final
{
public:
	using value_type = T;

public:
	explicit Sequencer() {}
	~Sequencer() {}

public:
	//MULTIPLE SEQUENCES
	void sortSequences(vector<T>& srcS, vector<SequenceInfo>& srcSI);
	bool scaleSequences(vector<T>& srcS, vector<SequenceInfo>& srcSI, const T& scaleFactor);
    bool denoiseSequences0(vector<T>& srcS, vector<SequenceInfo>& srcSI, const Comparator<T>& cmp);
	bool denoiseSequences1(vector<T>& srcS, vector<SequenceInfo>& srcSI, const T& delta);
	bool shiftSequences(vector<T>& srcS, const vector<SequenceInfo>& srcSI);
    
	bool getDeltaSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, vector<T>& dstS, vector<SequenceInfo>& dstSI);
	bool getReverseSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, vector<T>& dstS, vector<SequenceInfo>& dstSI);
	void getUniqueSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, const Comparator<T>& cmp, vector<SequenceInfo>& dstSI);
	bool getCountSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, const T slidingWindowSize, const T slideSize, const Comparator<T>& cmp, vector<size_t>& dstS, vector<SequenceInfo>& dstSI);

	//SINGLE SEQUENCE
	bool reverseSequence(T* __restrict__ s, const size_t length);
	bool smoothSequence(const T* __restrict__ s, const size_t length, const size_t neighbourhoodSize, float*& dst);

	//SLIDING WINDOW FUNCTIONS
	bool getFirstIndices(const T* __restrict__ s, const size_t length, const T slidingWindowSize, const T slideSize, vector<size_t>& indexVector);
	bool getFirstIndicesWithGaps(const T* __restrict__ s, const size_t length, const T slidingWindowSize, const T slideSize,  vector<size_t>& indexVector);
	bool getCounts(const T* __restrict__ s, const size_t length, const T slidingWindowSize, const T slideSize, const Comparator<T>& cmp, vector<size_t>& countVector);

	//STATS
	void getStatistics(const T* __restrict__ s, const size_t length, const T realLength, SequenceStatistics<T>& stat);	
};



///====================================================================================================
///@brief Sorts all elements of the sequenceInfo vector in descending order by length.
///
///@tparam T Sequence data type
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///====================================================================================================
template<typename T>
void Sequencer<T>::sortSequences(vector<T>& srcS, vector<SequenceInfo>& srcSI)
{
    auto sortFn = [&srcS](const SequenceInfo& a, const SequenceInfo& b) -> bool
                {
					if (a.length > b.length) return true;
					if (a.length < b.length) return false;

					//sequences have the same length = the same number of labels/sites
					const T* ptrA = &srcS[a.offset];
					const T* ptrB = &srcS[b.offset];
					for (uint32_t i=0; i<a.length; i++)
					{
						if (*ptrA == *ptrB) continue;
						return (*ptrA < *ptrB);
					}
					return true;
                };

    std::sort(srcSI.begin(), srcSI.end(), sortFn);
}

///====================================================================================================
///@brief Gets unique sequences and stores the resulting list into the vector dstSI
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@param cmp An equality comparator
///@param dstSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///====================================================================================================
template<typename T>
void Sequencer<T>::getUniqueSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, const Comparator<T>& cmp, vector<SequenceInfo>& dstSI)
{
	dstSI.reserve(srcSI.size());

	dstSI.emplace_back(srcSI[0]);
	for (size_t i=1; i<srcSI.size(); i++)
	{
		if (srcSI[i].length != srcSI[i-1].length)
		{
			dstSI.emplace_back(srcSI[i]);
			continue;
		}
		const T* ptrA = &srcS[srcSI[i].offset];
		const T* ptrB = &srcS[srcSI[i-1].offset];
		for (uint32_t j=0; j<srcSI[i].length; j++, ptrA++, ptrB++)
		{
			if (cmp.comparator0(*ptrA, *ptrB) != 0)
			{
				dstSI.emplace_back(srcSI[i]);
				break;			
			}
		}
	}

	dstSI.shrink_to_fit();    
}

///====================================================================================================
///@brief Scales sequences stored in the srcS 1D vector.
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@param scaleFactor Scale factor. Each sequence will be rescaled by this factor.
///@return bool TRUE if OK, otherwise FALSE
///====================================================================================================
template<typename T>
bool Sequencer<T>::scaleSequences(vector<T>& srcS, vector<SequenceInfo>& srcSI, const T& scaleFactor)
{
	ProgressBar progress0{std::clog, 70u, "Scaling sequences: "};
	uint32_t progressCounter = 0;
	float progressTotalCount = static_cast<float>(srcSI.size());
	
	T* ptrT = nullptr;

	for (const auto& si: srcSI)
	{
		ptrT = &srcS[si.offset];
		for (size_t i=0; i<si.length; i++, ptrT++)
		{
			*ptrT *= scaleFactor;
		}
		progressCounter++;
		progress0.write(progressCounter/progressTotalCount);
	}

	return true;
}

///====================================================================================================
///@brief Denoises sequences such that makes sequential pairwise comparison. If two values (I_0, I_1) are similar (see the Comparator), then the average is computed
/// and the average value replaces the two values withing the sequence. Then next two values (I_2, I_3, shift = 2) in the sequence are compared. If two values are NOT similar,
/// shift = 1 and other two values (I_1, I_2) are compared.
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@param cmp An equality comparator
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::denoiseSequences0(vector<T>& srcS, vector<SequenceInfo>& srcSI, const Comparator<T>& cmp)
{
	ProgressBar progress0{std::clog, 70u, "Removing noise: "};
	uint32_t progressCounter = 0;
	float progressTotalCount = static_cast<float>(srcSI.size());
	
	bool runAgain = true;

	vector<T> sAux;
	vector<T> sFinal;
	vector<SequenceInfo> siFinal;
	
	sAux.reserve(1<<12);
	sFinal.reserve(srcS.size());
	siFinal.reserve(srcSI.size());

	size_t finalOffset = 0;

	T* current = nullptr;
	T* next = nullptr;
	T* positionToSave = nullptr;

	size_t itemCounter = 0;
	size_t sAuxLength = 0;

	for (auto&& si: srcSI)
	{
		runAgain = true;
		sAuxLength = si.length;
		//sAux.items.copy(srcS.items, si.offset, si.length);
		pg::stdvectorutils::copyValues(srcS, si.offset, si.length, sAux);

		current = sAux.data();
		next = current+1;
		positionToSave = current;
		itemCounter = 0;

		while (runAgain)
		{
			runAgain = false;

			while(itemCounter < (sAuxLength-1))
			{
				if (cmp.comparator0(*current,*next)!=0)
				{
					*positionToSave = *current;
					current++;
					next++;
					itemCounter++;
				}
				else 
				{
					*positionToSave = (*current + *next) * 0.5;
					current+=2;
					next+=2;
					itemCounter+=2;
					runAgain = true;
				}
				positionToSave++;
			}

			if (current == &sAux[sAuxLength-1])	//Write the last element that was not compared.
			{
				*positionToSave = *current;
				itemCounter++;
				positionToSave++;
			}

			if (runAgain)
			{
				sAuxLength = positionToSave - sAux.data();
				current = sAux.data();
				next = current+1;
				positionToSave = current;
				itemCounter=0;
			}

			//cout << endl << sAux.items << endl;
		}

		siFinal.push_back(si);
		SequenceInfo& tmp = siFinal.back();
		tmp.length = positionToSave - sAux.data();
		tmp.offset = finalOffset;

		//for (auto&& x: sAux)
		for (uint32_t x = 0; x<tmp.length; x++)
		{
			sFinal.emplace_back(sAux[x]);
		}

		finalOffset+= tmp.length;
		progressCounter++;
		progress0.write(progressCounter/progressTotalCount);
	}

	sFinal.shrink_to_fit();
	srcS.swap(sFinal);
	
	siFinal.shrink_to_fit();
	srcSI.swap(siFinal);

	return true;
}

///====================================================================================================
///@brief Denoises sequences such that it makes sequential pairwise comparison up to total length of delta. If two or more values (I_i, I_i+1, ..., I_n) fall into delta interval
/// then they are similar (see the Comparator) and their average is computed. The average value replaces all parent values withing the sequence. 
/// Then other values (I_n+1, I_n+2, ...) in the sequence are compared.
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@param delta A number/length used to aggregate all values < than this length 
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::denoiseSequences1(vector<T>& srcS, vector<SequenceInfo>& srcSI, const T& delta)
{
	ProgressBar progress0{std::clog, 70u, "Removing noise: "};
	uint32_t progressCounter = 0;
	float progressTotalCount = static_cast<float>(srcSI.size());
	
	vector<T> sAux;
	vector<T> sFinal;
	vector<SequenceInfo> siFinal;

	sAux.reserve(1<<12);
	sFinal.reserve(srcS.size());
	siFinal.reserve(srcSI.size());

	size_t finalOffset = 0;

	T* last = nullptr;
	T* current = nullptr;
	T* next = nullptr;
	T* positionToSave = nullptr;

	bool runAgain = true;
	T sumDeltas = 0;
	T sumMarkers = 0;
	size_t sumCount = 0;

	for (auto&& si: srcSI)
	{
		runAgain = true;
		//sAux.items.copy(srcS.items, si.offset, si.length);
		pg::stdvectorutils::copyValues(srcS, si.offset, si.length, sAux);
		//last = sAux.items.back();
		//last = &(sAux.back());
		last = &sAux[sAux.size()- 1];

		while (runAgain)
		{
			runAgain = false;

			positionToSave = sAux.data();
			current = positionToSave;
			next = current + 1;

			sumCount = 1;
			sumMarkers = *current;
			sumDeltas = 0;

			while(current < last)
			{
				if ((sumDeltas + (*next - *current))<= delta)
				{
					sumDeltas += *next - *current;
					sumMarkers += *next;
					sumCount++;

					runAgain = true;					
					current++;
					next++;
				}
				else
				{
					*positionToSave = sumMarkers/sumCount;
					positionToSave++;
					current++;
					next++;
					sumCount = 1;
					sumMarkers = *current;
					sumDeltas = 0;					
				}
			}

			if (current == last)	//Write the last element that was not compared.
			{
				*positionToSave = sumMarkers/sumCount;
				positionToSave++;
			}

			if (runAgain)
			{
				last = positionToSave-1;
			}
		
			//cout << endl << sAux.items << endl;
		}

		siFinal.push_back(si);
		SequenceInfo& tmp = siFinal.back();
		tmp.length = positionToSave - sAux.data();
		tmp.offset = finalOffset;

		//for (auto&& x: sAux)
		for (uint32_t x = 0; x<tmp.length; x++)
		{
			sFinal.emplace_back(sAux[x]);
		}


		finalOffset+= tmp.length;
		progressCounter++;
		progress0.write(progressCounter/progressTotalCount);
	}

	sFinal.shrink_to_fit();
	srcS.swap(sFinal);
	
	siFinal.shrink_to_fit();
	srcSI.swap(siFinal);

	return true;
}

///====================================================================================================
///@brief Shifts sequences such that the first label/site position is set to 0.
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@return bool TRUE if OK, otherwise FALSE
///====================================================================================================
template<typename T>
bool Sequencer<T>::shiftSequences(vector<T>& srcS, const vector<SequenceInfo>& srcSI)
{
	T *ptrT = nullptr;
	for (auto&& si: srcSI)
	{
		ptrT = &srcS[si.offset];
		const T first = *ptrT;
		for (size_t i=0; i<si.length; i++, ptrT++)
		{
			*ptrT -= first;
		}
	}
	return true;
}

///====================================================================================================
///@brief Goes through the src sequences and creates a new such that dstS[i] = srcS[i+1] - srcS[i]. Moreover, it updates vector of SequenceInfos,
///		  because the resulting sequences are shorter (N-1)
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@param dstS Vector of all DELTA sequences stored in 1D array of T datatype. 
///@param dstSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the dstS vector.
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::getDeltaSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, vector<T>& dstS, vector<SequenceInfo>& dstSI)
{
	ProgressBar progress0{std::clog, 70u, "Creating delta sequences: "};
	uint32_t counter = 0;
	float totalCount = static_cast<float>(srcSI.size());

	dstS.reserve(srcS.size()-1);
	dstSI.reserve(srcSI.size());
	uint32_t offset = 0;
	for (const auto& si: srcSI)
	{
		for (size_t i=si.offset+1; i<si.offset+si.length; i++)
		{
			dstS.emplace_back(srcS[i]-srcS[i-1]);
		}
		dstSI.emplace_back(SequenceInfo());
		SequenceInfo& tmp = dstSI.back();
		tmp.length = si.length-1;
		tmp.id = si.id;
		tmp.offset = offset;
		offset += tmp.length;

		counter++;
		progress0.write(counter/totalCount);
	}
	dstS.shrink_to_fit();
	dstSI.shrink_to_fit();
	return true;
}

///====================================================================================================
///@brief Creates reverse sequences.
///
///@tparam T 
///@param srcS Vector of all sequences stored in 1D array of T datatype. 
///@param srcSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the srcS vector.
///@param dstS Vector of all DELTA sequences stored in 1D array of T datatype. 
///@param dstSI Vector of SequenceInfos. Each SequenceInfo holds the storage information of i-th sequence in the dstS vector.
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::getReverseSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, vector<T>& dstS, vector<SequenceInfo>& dstSI)
{
	ProgressBar progress0{std::clog, 70u, "Creating reverse sequences: "};
	uint32_t counter = 0;
	float totalCount = static_cast<float>(srcSI.size());

	dstS.reserve(srcS.size());
	dstSI.reserve(srcSI.size());
	T position = 0;
	for (const auto& si: srcSI)
	{
		position = 0;
		size_t i=0;
		while(i<si.length)
		{
			dstS.emplace_back(position);
			if (++i>=si.length) break;
			position += srcS[si.offset+si.length-i]-srcS[si.offset+si.length-i-1];
		}
		dstSI.emplace_back(si);
		counter++;
		progress0.write(counter/totalCount);
	}
	dstS.shrink_to_fit();
	dstSI.shrink_to_fit();
	return true;
}

template<typename T>
bool Sequencer<T>::getCountSequences(const vector<T>& srcS, const vector<SequenceInfo>& srcSI, const T slidingWindowSize, const T slideSize, const Comparator<T>& cmp, vector<size_t>& dstS, vector<SequenceInfo>& dstSI)
{
	ProgressBar progress0{std::clog, 70u, "Creating count sequences: "};
	uint32_t counter = 0;
	float totalCount = static_cast<float>(srcSI.size());

	size_t dstLength = 0;
	for (auto&&si: srcSI)
	{
		dstLength += pg::slidingwindowutils::getNumberOfWindows(slidingWindowSize, slideSize, static_cast<T>(srcS[si.offset+si.length-1]));
	}

	dstS.reserve(dstLength);
	dstSI.reserve(srcSI.size());
	
	vector<size_t> countVector;
	dstLength = 0;
	for (const auto& si: srcSI)
	{
		getCounts(&srcS[si.offset], si.length, slidingWindowSize, slideSize, cmp, countVector);
		std::copy(countVector.begin(), countVector.end(), std::back_inserter(dstS));

		dstSI.emplace_back(SequenceInfo());
		SequenceInfo& tmp = dstSI.back();
		tmp.offset = static_cast<uint32_t>(dstLength);
		tmp.length = static_cast<uint32_t>(countVector.size());
		tmp.id = si.id;
		dstLength += countVector.size();

		countVector.clear();
		counter++;
		progress0.write(counter/totalCount);
	}
	dstS.shrink_to_fit();
	dstSI.shrink_to_fit();
	return true;	
}



///====================================================================================================
///@brief Creates a reverse sequence from a given one.
///
///@tparam T 
///@param s An array of T values representing a single sequence.
///@param length The length of the sequence.
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::reverseSequence(T* __restrict__ s, const size_t length)
{
	if (length <=1) return false;
	for (size_t i = 0; i<length/2; i++)
	{
		swap(s[i], s[length-1-i]);
	}

	T prev = static_cast<T>(0);
	T curr = prev;
	swap(prev,s[0]);
	for (size_t i = 1; i<length; i++)
	{
		curr = s[i];
		s[i] = s[i-1] + prev - curr;
		prev= curr;
	}
	return true;
}

///====================================================================================================
///@brief NoOverlappingWindows version => Creates an index vector, where each element is an index of 
///		  the first s_element that falls into a give sliding window of the size of windowSize and slideSize = windowSize. 
///		  If there is no s_element that falls into a give sliding window, the index value is set later using a consecutive s_element (see the while loop).
///		  The index vector is an array of indices of type size_t. 
///		  E.g.  	   s = {26,38,61,100} and sliding windowSize = 25
///				sIndices = {0, 1, 2, 3}
/// 	  For |indexVector| = 5 see getNumberOfWindows()
///		  indexVector = {0,0,2,3,3}
///@tparam T 
///@param s An array of T values representing a single sequence having shifted labels. The first label position = 0 and the last label position = sequenceLength.
///@param length The length of S.
///@param slidingWindowSize Size of the sliding window. O for non overlapping windows.
///@param slideSize Shift size of the the sliding window.
///@param indexVector Resulting vector.
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::getFirstIndices(const T* __restrict__ s, const size_t length, const T slidingWindowSize, const T slideSize, vector<size_t>& indexVector)
{
	if (length<2) return false;
	if ((slideSize < 0) || (slidingWindowSize<0) || (slidingWindowSize < slideSize)) return false;

	//No overlapping windows
	if (slideSize == T(0))
	{
		T sequenceLength = s[length-1];
		size_t noWindows = pg::slidingwindowutils::getNumberOfWindows(slidingWindowSize, slidingWindowSize, sequenceLength);
		indexVector.resize(noWindows, 0);

		size_t idx = 0;
		size_t wid = 0;

		for (size_t i=0; i<length; i++)
		{
			idx = static_cast<size_t>(s[i]/slidingWindowSize);
			while(wid <= idx)
			{
				indexVector[wid++] = i;
			}
		}
		return true;
	}
	return false;
}


///====================================================================================================
///@brief NoOverlappingWindows version => Creates an index vector, where each element is an index of 
///		  the first s_element that fall into a give sliding window of the size of windowSize and slideSize = windowSize. 
/// 	  The index vector is an array of size_t indices. 
///		  E.g.  s = {26,38,61,100} and sliding windowSize = 25
/// 	  For |indexVector|=5 see getNumberOfWindows()
///		  indexVector = {MAX=GAP,0,2, MAX=GAP, 3}
///		  std::numeric_limits<size_t>::max() indicates GAP = no s_element has fallen into the sliding window.
///@tparam T 
///@param s An array of T values representing a single sequence having shifted labels. The last label position = sequenceLength.
///@param length The length of S.
///@param slidingWindowSize Size of the sliding window. O for non overlapping windows.
///@param slideSize Shift size of the the sliding window.
///@param indexVector Resulting vector.
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::getFirstIndicesWithGaps(const T* __restrict__ s, const size_t length,  const T slidingWindowSize, const T slideSize, vector<size_t>& indexVector)
{
	if (length<1) return false;
	if ((slideSize < 0) || (slidingWindowSize<0) || (slidingWindowSize < slideSize)) return false;

	//No overlapping windows
	if (slideSize == T(0))
	{
		T sequenceLength = s[length-1];

		size_t noWindows = pg::slidingwindowutils::getNumberOfWindows(slidingWindowSize, slidingWindowSize, sequenceLength);
		size_t maxIdx = std::numeric_limits<size_t>::max();
		indexVector.resize(noWindows, maxIdx);

		size_t wid = 0;

		for (size_t i=0; i<length; i++)
		{
			wid = static_cast<size_t>(s[i]/slidingWindowSize);
			if (i < indexVector[wid])
			{
				indexVector[wid] = i;
			}
		}
		return true;
	}
	return false;
}


///====================================================================================================
///@brief Creates an count vector.  The count vector is an array of counters. 
///		  Every element represents the number of s_elements that fall into a give window of the size of windowSize.
///		  E.g.  Let comparator's m_epsilon is 0.0
///						  s = {26,38,61,100} and windowSize = 25
/// 	  For |indexVector| = 5 see getNumberOfWindows()
///		  		indexVector = {0,2,1,0,1}
///@tparam T 
///@param s An array of T values representing a single sequence having shifted labels. The last label position = sequenceLength.
///@param length The length of S.
///@param slidingWindowSize Size of the sliding window. O for non overlapping windows.
///@param slideSize Shift size of the the sliding window.
///@param cmp An equality comparator
///@param indexVector Resulting vector.
///@return bool 
///====================================================================================================
template<typename T>
bool Sequencer<T>::getCounts(const T* __restrict__ s, const size_t length, const T slidingWindowSize, const T slideSize, const Comparator<T>& cmp, vector<size_t>& countVector)
{
	if (length<1) return false;

	T sequenceLength = s[length-1];

	//No overlapping windows
	if (slideSize == T(0))
	{
		size_t noWindows = pg::slidingwindowutils::getNumberOfWindows(slidingWindowSize, slidingWindowSize, sequenceLength);
		countVector.resize(noWindows, 0);

		size_t wid = 0;
		for (size_t i=0; i<length; i++)
		{
			wid = static_cast<size_t>(s[i]/slidingWindowSize);
			countVector[wid]++;
		}
		return true;
	}
	else //overlapping windows
	{
		size_t noWindows = pg::slidingwindowutils::getNumberOfWindows(slidingWindowSize, slideSize, sequenceLength);
		countVector.resize(noWindows, 0);

		size_t wid = 0;
		T startPos = 0;
		T endPos = 0;
		for (size_t i=0; i<length; i++)
		{
			wid = pg::slidingwindowutils::getFirstWindowIdx(slidingWindowSize, slideSize, s[i]);
			startPos = static_cast<T>(wid) * slideSize;
			endPos = startPos + slidingWindowSize;

			if ((cmp.comparator0(startPos,s[i]) <= 0)&&(cmp.comparator0(s[i], endPos) <= 0)&&(wid<noWindows))
			{
				countVector[wid]++;
				startPos = static_cast<T>(++wid) * slideSize;
				endPos = startPos + slidingWindowSize;
			}

			// while ((startPos<=s[i])&&(s[i] < endPos)&&(wid<noWindows))
			// {
			// 	countVector[wid]++;
			// 	startPos = static_cast<T>(++wid) * slideSize;
			// 	endPos = startPos + slidingWindowSize;
			// }
		}
		return true;
	}
}


template<typename T>
bool Sequencer<T>::smoothSequence(const T* __restrict__ s, const size_t length, const size_t neighbourhoodSize, float*& dst)
{
	dst = static_cast<float*>(::operator new(length * sizeof(float)));

	float attenuationFactor = 1.0f/(neighbourhoodSize+1);
	float value = 0.0f;
	
	for (size_t i=0; i<length; i++)
	{
		dst[i] = static_cast<float>(s[i]);
		if (s[i] > 0)
		{
			for (size_t j=0; j<neighbourhoodSize; j++)
			{
				value = (neighbourhoodSize-j) * attenuationFactor *  s[i];
				
				//update neighbors from the current position to the RIGHT
				if ((i+1+j)<length)
				{
					dst[i+1+j] += value;
				}

				//update neighbors from the current position to the LEFT
				if (i >= (1+j))
				{
					dst[i-1-j] += (neighbourhoodSize-j) * attenuationFactor *  s[i];
				}	
			}
		}
	}
	return true;
}


template<typename T>
void Sequencer<T>::getStatistics(const T* __restrict__ s, const size_t length, const T realLength, SequenceStatistics<T>& stat)
{
	memset(&stat, 0, sizeof(SequenceStatistics<T>));				

	const T partSize = realLength / ssNoLeaves;
	uint32_t level = ssDepth - 1;									//current level of the binary tree
	uint32_t currentNID = (1 << level) - 1; 						//ID of the first node on a given level of the binary three
	uint32_t i=1;
	T delta;
	uint32_t partID = 0;

	stat.noSites[currentNID] = 1;
	stat.minDelta = std::numeric_limits<T>::max();
	stat.length = 0;
	stat.norm = 0;

	while (i < length)
	{
		delta = s[i] - s[i-1];

		if ((stat.length + delta) > realLength) break;

		partID = static_cast<uint32_t>((stat.length + delta)/partSize);

		if (partID >= ssNoLeaves) partID = ssNoLeaves -1;

		stat.length+=delta;
		stat.noSites[currentNID+partID]++;

		if (delta < stat.minDelta) stat.minDelta = delta;
		if (delta > stat.maxDelta) stat.maxDelta = delta;
		i++;
	}

	//Build the upper levels of the binary tree
	uint32_t prevNID;
	while (level>0)
	{
		prevNID = currentNID;
		--level;
		currentNID = (1 << level) - 1;

		for (i=0; i<(1<<level); i++)
		{
			stat.noSites[currentNID + i] = stat.noSites[prevNID + i*2] + stat.noSites[prevNID + i*2 + 1];
			stat.norm += stat.noSites[currentNID + i] * stat.noSites[currentNID + i];
		}
	}

	if (stat.noSites[0] > 1)
	{
		stat.avgDelta = stat.length / stat.noSites[0];
		stat.norm += stat.avgDelta * stat.avgDelta;
		stat.norm += stat.length * stat.length;
		stat.norm += stat.maxDelta * stat.maxDelta;
		stat.norm += stat.minDelta * stat.minDelta;
		stat.norm = std::sqrt(stat.norm);
	}
	else
	{
		stat.minDelta = 0;
		stat.maxDelta = 0;
		stat.avgDelta = 0;
		stat.length = 0;
		stat.norm = 0;
	}
	
}