///====================================================================================================
///@file      bioStructures.h
///@author    Petr Gajdoš (petr.gajdos@vsb.cz)
///@brief 	  This header file defines all required data types and typle aliases
///@version   1.0
///@date      2020-01-12
///
///@copyright Copyright (c) 2020
///
///====================================================================================================

#pragma once

#include <cstddef>
#include <cstdint>
#include <iomanip>
#include <fstream>
#include <string>
#include <vector>
#include <tuple>

#include <utils/sfinae.h>

#define USE_OMP


///====================================================================================================
///@brief Helper structures to store records in Row or Column manner
///====================================================================================================
// template<typename... Ts>
// struct VofT : public std::vector<std::tuple<Ts...>>{};

// template<typename... Ts>
// struct TofV : public std::tuple<std::vector<Ts>...>{};

// template<typename... Ts>
// struct Typelist
// {
// static constexpr size_t count{sizeof...(Ts)};
// };

// template<class... ARGS>
// std::tuple<std::vector<ARGS>... > typelist_helper(Typelist<ARGS...>);

// template<class T>
// using recordset = decltype(typelist_helper(std::declval<T>()));

// using customRecord = Typelist<int, float, double>;
// recordset<customRecord> r;

template<typename... Ts> using VofT = std::vector<std::tuple<Ts...>>;
template<typename... Ts> using TofV = typename std::tuple<std::vector<Ts>...>;

///====================================================================================================
///@brief This tuples are space ineffective because of bad order of inner members (interleaved strings, etc.). 
// 		  But it has to fit physical data structure of input files to simplify general loding and parsing.
///====================================================================================================
using xmap_v02 = std::tuple
<
	std::uint32_t,	// XmapEntryID
	std::uint32_t,	// QryContigID
	std::uint32_t,	// RefContigID
	double,			// QryStartPos
	double,			// QryEndPos
	double,			// RefStartPos
	double,			// RefEndPos
	std::string,	// Orientation
	float,			// Confidence
	std::string,	// HitEnum
	double,			// QryLen
	double,			// RefLen
	std::int32_t,	// LabelChannel
	std::string		// Alignment
>;

using smap_v08 = std::tuple
<
	std::uint32_t,	// SmapEntryID
	std::uint32_t,	// QryContigID
	std::uint32_t,	// RefcontigID1
	std::uint32_t,	// RefcontigID2
	double,			// QryStartPos
	double,			// QryEndPos
	double,			// RefStartPos
	double,			// RefEndPos
	float,			// Confidence
	std::string,	// Type
	std::uint32_t,	// XmapID1
	std::uint32_t,	// XmapID2
	std::int32_t,	// LinkID
	std::uint32_t,	// QryStartIdx
	std::uint32_t,	// QryEndIdx
	std::uint32_t,	// RefStartIdx
	std::uint32_t,	// RefEndIdx
	std::string,	// Zygosity
	std::int32_t,	// Genotype
	std::int32_t,	// GenotypeGroup
	float,			// RawConfidence
	float,			// RawConfidenceLeft
	float,			// RawConfidenceRight
	float,			// RawConfidenceCenter
	float			// SVsize
>;

using smap_v09 = std::tuple
<
	std::uint32_t,	// SmapEntryID
	std::uint32_t,	// QryContigID
	std::uint32_t,	// RefcontigID1
	std::uint32_t,	// RefcontigID2
	double,			// QryStartPos
	double,			// QryEndPos
	double,			// RefStartPos
	double,			// RefEndPos
	float,			// Confidence
	std::string,	// Type
	std::uint32_t,	// XmapID1
	std::uint32_t,	// XmapID2
	std::int32_t,	// LinkID
	std::uint32_t,	// QryStartIdx
	std::uint32_t,	// QryEndIdx
	std::uint32_t,	// RefStartIdx
	std::uint32_t,	// RefEndIdx
	std::string,	// Zygosity
	std::int32_t,	// Genotype
	std::int32_t,	// GenotypeGroup
	float,			// RawConfidence
	float,			// RawConfidenceLeft
	float,			// RawConfidenceRight
	float,			// RawConfidenceCenter
	float,			// SVsize
	float,			// SVfreq
	std::string,	// orientation
	std::string,	// Sample
	std::string,	// Algorithm
	std::uint32_t,	// Size
	float,			// Present_in_%_of_BNG_control_samples
	float,			// Present_in_%_of_BNG_control_samples_with_the_same_enzyme
	std::string,	// Fail_assembly_chimeric_score
	std::string,	// OverlapGenes
	std::string,	// NearestNonOverlapGene
	float,			// NearestNonOverlapGeneDistance
	std::string,	// PutativeGeneFusion
	std::string,	// Found_in_self_molecules
	std::uint32_t	// Self_molecule_count
>;

using cmap_v01 = std::tuple
<
	std::uint32_t,	// CMapId
	double,			// ContigLength
	std::uint32_t,	// NumSites
	std::uint32_t,	// SiteID
	std::uint32_t,	// LabelChannel
	double,			// Position
	float,			// StdDev
	float,			// Coverage
	float,			// Occurrence
	float,			// ChimQuality
	float,			// SegDupL
	float,			// SegDupR
	float,			// FragileL
	float,			// FragileR
	float,			// OutlierFrac
	float			// ChimNorm
>;

using cmap_v02 = std::tuple
<
	std::uint32_t,	// CMapId
	double,			// ContigLength
	std::uint32_t,	// NumSites
	std::uint32_t,	// SiteID
	std::uint32_t,	// LabelChannel
	double,			// Position
	float,			// StdDev
	float,			// Coverage
	float,			// Occurrence
	float,			// ChimQuality
	float,			// SegDupL
	float,			// SegDupR
	float,			// FragileL
	float,			// FragileR
	float,			// OutlierFrac
	float,			// ChimNorm
	std::uint64_t	// Mask
>;

using cmap_v02q = std::tuple
<
	std::uint32_t,	// CMapId
	double,			// ContigLength
	std::uint32_t,	// NumSites
	std::uint32_t,	// SiteID
	std::uint32_t,	// LabelChannel
	double,			// Position
	float,			// StdDev
	float,			// Coverage
	float,			// Occurrence
	float,			// ChimQuality
	float,			// SegDupL
	float,			// SegDupR
	float,			// FragileL
	float,			// FragileR
	float,			// OutlierFrac
	float,			// ChimNorm
	std::uint64_t,	// Mask
	float,			// GmeanSNR
	float			//lnSNRsd
>;

using cmap_v02q_sub0 = std::tuple
<
	std::uint32_t	// CMapId
>;

using bnx_v13_0 = std::tuple
<
	std::uint8_t, 			// LabelChannel
	std::uint32_t,			// MoleculeID
	double,					// Length
	double,					// AvgIntensity
	double,					// SNR
	std::uint32_t,			// NumberofLabels = N
	std::uint32_t,			// OriginalMoleculeId
	std::uint8_t,			// ScanNumber
	std::uint8_t,			// ScanDirection
	std::string,			// ChipId
	std::uint8_t,			// Flowcell
	std::uint32_t,			// RunId
	std::uint32_t			// GlobalScanNumber	
>;

using bnx_v13_1 = std::pair
<
	std::uint8_t, 			// LabelChannel
	std::vector<double> 	// LabelPositions[N+1] - the last element is total length = bnx_v13_0[Length]
>;

using bnx_v13_Q = std::pair	// Quality Score
<
	std::string, 			// QualityScoreID
	std::vector<float> 		// QualityScores[N]
>;

struct bnx_v13
{
	bnx_v13_0 base;
	bnx_v13_1 positions;
	bnx_v13_Q qx11;			// Quality Score QX11: Label SNR for channel 1
	bnx_v13_Q qx12;			// Quality Score QX12: Label Intensity for channel 1
};

using namespace pg::sfinae;
define_has_member(base);	//Check in code in compile-time ... static_assert(has_member(SOME_TYPE, base), "SOME_TYPE must contain base member");
define_has_member(positions);
define_has_member(qx11);
define_has_member(qx12);


using XMAPPair = std::pair
<
	std::uint32_t, 	// query site/marker ID
	std::uint32_t	// reference site/marker ID
>;

using XMAPTuple = std::tuple
<
	std::uint32_t,	// QryContigID
	std::uint32_t,	// RefContigID
	std::uint32_t,	// query site/marker ID
	std::uint32_t	// reference site/marker ID
>;

using idMapping = std::pair		// e.g. Reference contig to Query contig mapping
<
	std::uint32_t, 				// src ID
	std::vector<std::uint32_t>	// mapping IDs
>;

enum class EBionanoFileType : uint8_t
{	
	xmap_v02,
	smap_v08,
	smap_v09,
	cmap_v01,
	cmap_v02,
	cmap_v02q,
	cmap_v02q_sub0,
	bnx_v13,
	UNDEFINED
};

struct SequenceInfo
{
    std::uint32_t id;			//DO NOT CHANGE ORDER OF VARIABLES BECAUSE OF USE OF {} CONSTRUCTS in the code !!!
    std::uint32_t offset;
    std::uint32_t length;

    friend std::ostream& operator <<(std::ostream& s, const SequenceInfo& si)
	{
		s << si.id << " " << si.offset << " " << si.length << std::endl;
		return s;
	}
};

struct ImageAlignmentMetricData
{
	std::uint32_t rIdx;			//The index of A sequence
	std::uint32_t rOffset;		//Offset on A where the matching started
	float fitness;

	ImageAlignmentMetricData(const ImageAlignmentMetricData&) = default;
	ImageAlignmentMetricData(ImageAlignmentMetricData&&) = default;

	ImageAlignmentMetricData& operator=(const ImageAlignmentMetricData&) = default;			//Forcing a copy assignment operator to be generated by the compiler
	ImageAlignmentMetricData& operator=(ImageAlignmentMetricData&&) = default;				//Forcing a move assignment operator to be generated by the compiler

	friend bool operator<(const ImageAlignmentMetricData& l, const ImageAlignmentMetricData& r)
	{
		return l.fitness < r.fitness;
	}
	friend bool operator>(const ImageAlignmentMetricData& l, const ImageAlignmentMetricData& r)
	{
		return l.fitness > r.fitness;
	}

	friend std::ostream& operator <<(std::ostream& s, const ImageAlignmentMetricData& m)
	{
		s  << "[" << m.rIdx << "|" << m.rOffset << "|" << std::setprecision(3) << m.fitness << "] ";
		return s;
	}
};


template<typename T>
struct AlignmentMetricData
{
	std::uint32_t a;			//The number of marks of A not found in B
	std::uint32_t b;			//The number of marks of B not found in A
	std::uint32_t c;			//The number of matches = common labels, markers
	std::uint32_t lcp;			//The largest common part = the length of the largest cointinuous part of exact matches C
	std::size_t aOffset;		//Offset on A where the matching started
	std::size_t bOffset;		//Offset on B where the matching started
	T bPositionOffset;			//The position offset of virtual labels of B
	std::uint32_t aIdx;			//The index of A sequence
	bool direction;				//TRUE in case of Left->Right direction, otherwise FALSE.

	inline std::uint32_t count() const { return c + a + b; }
	inline std::uint32_t countAB() const { return a + b; }

	AlignmentMetricData(const AlignmentMetricData<T>&) = default;
	AlignmentMetricData(AlignmentMetricData<T>&&) = default;

	AlignmentMetricData<T>& operator=(const AlignmentMetricData<T>&) = default;			//Forcing a copy assignment operator to be generated by the compiler
	AlignmentMetricData<T>& operator=(AlignmentMetricData<T>&&) = default;				//Forcing a move assignment operator to be generated by the compiler

	friend bool operator<(const AlignmentMetricData<T>& l, const AlignmentMetricData<T>& r)
    {
		if (l.c > r.c) return true;
		else if (l.c == r.c)
		{
			if (l.b < r.b) return true;
			else if (l.b == r.b) 
			{
				return (l.a < r.a);
			}
		}
		return false;

		// if (l.lcp > r.lcp) return true;
		// else if (l.lcp == r.lcp)
		// {
		// 	if (l.c > r.c) return true;
		// 	else if (l.c == r.c) 
		// 	{
		// 		if (l.countAB() < r.countAB()) return true;
		// 		else if (l.countAB() == r.countAB()) 
		// 		{
		// 			if (l.b < r.b) return true;
		// 		}
		// 	}
		// }
		// return false;
    }

	friend bool operator>(const AlignmentMetricData<T>& l, const AlignmentMetricData<T>& r)
    {
		if (l.c < r.c) return true;
		else if (l.c == r.c)
		{
			if (l.b > r.b) return true;
			else if (l.b == r.b) 
			{
				return (l.a > r.a);
			}
		}
		return false;


		// if (l.lcp < r.lcp) return true;
		// else if (l.lcp == r.lcp)
		// {
		// 	if (l.c < r.c) return true;
		// 	else if (l.c == r.c) 
		// 	{
		// 		if (l.countAB() > r.countAB()) return true;
		// 		else if (l.countAB() == r.countAB()) 
		// 		{
		// 			if (l.b > r.b) return true;
		// 		}
		// 	}
		// }
		// return false;
	}

	friend std::ostream& operator <<(std::ostream& s, const AlignmentMetricData& m)
	{
		s 	<< m.lcp << "\t" 
			<< m.c << "\t" 
			<< m.a << "\t" 
			<< m.b << "\t" 
			<< ((m.direction) ? "true" : "false") << "\t" 
			<< m.aIdx << "\t" 
			<< m.aOffset << "\t" 
			<< m.bOffset << "\t" 
			<< m.bPositionOffset << std::endl;
		return s;
	}
};

struct AlignmentData
{
	std::uint32_t sid = 0;  	//sequence id
	std::uint32_t cid = 0;  	//contig id 
	std::uint32_t rsid = 0;		//reference sequence id
	std::size_t	 sStart=0;		//start matching position in the source sequence
	std::size_t	 sEnd=0;		//end matching position in the source sequence
	std::size_t	 rsStart=0;		//start matching position in the reference sequence
	std::size_t	 rsEnd=0;		//end matching position in the reference sequence

	AlignmentData(const AlignmentData&) = default;
	AlignmentData(AlignmentData&&) = default;

    AlignmentData& operator=(const AlignmentData& rhs) = default;
    AlignmentData& operator=(AlignmentData&& rhs) noexcept = default;

	friend std::ostream& operator <<(std::ostream& s, const AlignmentData& ad)
	{
		s << ad.sid << "\t" 
		<< ad.cid << "\t"
		<< ad.rsid << "\t"
		<< ad.sStart << "\t"
		<< ad.sEnd << "\t"
		<< ad.rsStart << "\t"
		<< ad.rsEnd << std::endl;
		return s;
	}
};


constexpr uint32_t ssDepth = 4;							//Depth of the binary tree of statistics
constexpr uint32_t ssNoLeaves = 1 << (ssDepth-1);		//Number of leaves of the binary tree of statistics
constexpr uint32_t ssNoNodes = (1 << ssDepth) - 1;		//Number of nodes of the binary tree of statistics

template<typename T>
struct SequenceStatistics
{
	uint32_t noSites[ssNoNodes];				//Sequence is binary divided into parts/nodes. Every node consists of the number of sites in part
	T minDelta;									//minimum difference between two consecutive sites
	T maxDelta;									//maximum difference between two consecutive sites
	T avgDelta;									//average difference between sites
	T length;									//length = (positionOfLastSite - positionOfFirstSite) in the target statistics area
	T norm;										//the norm of the whole SequenceStatistic vector
};