////////////////////////////////////////////////////////////////////////////////////////////////////
// file: COMPARATOR.h
// version: 1.0
// author: Petr Gajdos
// summary:	Templated class to manage COMPARISONS
// 
// NOTICE: All vectors/sets must contain UNIQUE items and must be ALLOCATED in advance.
//		   It is supposed that the |M[i]-M[i-1]|<epsilon in both sets
//
////////////////////////////////////////////////////////////////////////////////////////////////////
#pragma once
#include <cstddef>

namespace pg
{

	template<typename T> 
	class Comparator final
	{
	public:
		using value_type = T;
		using ComparatorFunctionPtr = char(*)(const T&, const T&);

	private:
		const T m_epsilon;

	private:
		size_t intersectionSize(const T* A, const T* B, const size_t lengthA, const size_t lengthB, const ComparatorFunctionPtr fptr) const;
		size_t unionSize(const T* A, const T* B, const size_t lengthA, const size_t lengthB, const ComparatorFunctionPtr fptr) const;

	public:
		explicit Comparator(const T epsilon) : m_epsilon(epsilon) { }
		~Comparator() {}

	public:
		inline const T getEpsilon() const {return m_epsilon;}

		char comparator0(const T& a, const T& b) const;
		float jaccard(const T* A, const T* B, const size_t lengthA, const size_t lengthB, const ComparatorFunctionPtr fptr) const;
	};


	//Returns 0 iff |a-b|<= m_epsilon, 1 iff ((b+m_epsilon)<a), -1 iff ((a+m_epsilon)<b)
	template<typename T>char Comparator<T>::comparator0(const T& a, const T& b) const
	{
		if (a < b) return ((a+m_epsilon)<b) ? -1 : 0;
		else if (b < a) return ((b+m_epsilon)<a) ? 1 : 0;
		return 0;
	}

	template<typename T> size_t Comparator<T>::intersectionSize(const T* A, const T* B, const size_t lengthA, const size_t lengthB, const ComparatorFunctionPtr fptr) const
	{
		const T* a = A;
		const T* b = B;
		size_t result = 0;
		size_t i =0;
		size_t j =0;

		while ((i<lengthA)&&(j<lengthB))
		{
			switch (fptr(*a, *b))
			{
				case -1: a++; i++; break;
				case 0: a++; b++; i++; j++; result++; break;
				case 1: b++; j++; break;
				default: break;
			}
		}

		return result;
	}

	template<typename T> size_t Comparator<T>::unionSize(const T* A, const T* B, const size_t lengthA, const size_t lengthB, const ComparatorFunctionPtr fptr) const
	{
		const T* a = A;
		const T* b = B;
		size_t result = 0;
		size_t i =0;
		size_t j =0;

		while ((i<lengthA)&&(j<lengthB))
		{
			switch (fptr(*a, *b))
			{
				case -1: a++; i++; break;
				case 0: a++; b++; i++; j++; break;
				case 1: b++; j++; break;
				default: break;
			}
			result++;
		}

		while (i<lengthA)
		{
			result++;
			i++;
		}

		while (j<lengthB)
		{
			result++;
			j++;
		}

		return result;
	}

	template<typename T>float Comparator<T>::jaccard(const T* A, const T* B, const size_t lengthA, const size_t lengthB, const ComparatorFunctionPtr fptr) const
	{
		float capSize = static_cast<float>(intersectionSize(A,B,lengthA, lengthB, fptr));
		return capSize / (lengthA+lengthB-capSize);
	}

}