#pragma once

#include <experimental/filesystem>

#include "defs/multiplatformDefs.h"

#include <fstream>
#include <sstream>
#include <string_view>

namespace fs = std::experimental::filesystem;

namespace pg::fileutils
{
    inline bool createOFS(std::ofstream &ofs, const std::string& fileName)
    {
        try 
        {
            if (ofs.is_open())
            {
                ofs.flush();
                ofs.close();
            }
            ofs = std::ofstream(fileName);
            if (!ofs.is_open())
            {
                std::cout << "Can not open the file: %d\n" <<  __LINE__ << std::endl;
                return false;
            }
            return true;
        }
        catch (const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    inline bool getBuffer(const std::string& fileName, char *&data, std::streampos &length)
    {
        try
        {
            std::ifstream f(fileName);
            if (!f.is_open())
            {
                return false;
            }

            f.seekg(0, std::ios::end);
            length = f.tellg();

            if (data != nullptr)
            {
                delete[] data;
            }
            data = static_cast<char *>(::operator new(static_cast<size_t>(length) + 1));     //Add one more extra byte to awoid memory sanitizer error when parsing string by e.g. string_view

            f.seekg(0, std::ios::beg);
            f.read(data, length);
            f.close();

            data[length] = '\0';                                        //Fill last extra byte
            return true;
        }
        catch(const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    inline bool getStringstream(const std::string& fileName, std::stringstream& ss)
    {
        try
        {
            std::ifstream f(fileName);
            if (!f.is_open())
            {
                return false;
            }

            f.seekg(0, std::ios::end);
            std::streampos length = f.tellg();

            char *data = static_cast<char *>(::operator new(length));

            f.seekg(0, std::ios::beg);
            f.read(data, length);
            f.close();
            
            ss.rdbuf()->pubsetbuf(data, length);

            return true;
        }
        catch(const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    inline void removePath(const std::string& path)
    {
        fs::path p = path;
        std::uintmax_t n = fs::remove_all(path);
        std::cout << "Deleted " << n << " files or directories\n" << std::endl;
    }

    inline int createPath( mode_t mode, const std::string& rootPath, const std::string& path)
    {
        struct stat st;
        const char sep = '/';

        std::string_view sw = rootPath.back() == '/' ? std::string_view(rootPath.c_str(), rootPath.length() - 1) : std::string_view(rootPath);
        std::string rPath = std::string(sw);

        std::string::const_iterator pathStart = path.front() == '/' ? path.begin()+1 : path.begin();
        std::string::const_iterator pathEnd = path.end();

        for(std::string::const_iterator iter = pathStart ; iter < pathEnd; )
        {
            std::string::const_iterator newIter = std::find(iter, pathEnd, sep);
            std::string newPath = (rPath == "") ? "/" + std::string(pathStart, newIter) : rPath + "/" + std::string(pathStart, newIter);

            if( stat( newPath.c_str(), &st) != 0)
            {           
                if( createDir( newPath.c_str(), mode) != 0 && errno != EEXIST )
                {
                    std::cout << "cannot create folder [" << newPath << "] : " << strerror(errno) << std::endl;
                    return -1;
                }
            }
            else
                if( !S_ISDIR(st.st_mode) )
                {
                    errno = ENOTDIR;
                    std:: cout << "path [" << newPath << "] not a dir " << std::endl;
                    return -1;
                }
                //  else
                //      std::cout << "path [" << newPath << "] already exists " << std::endl;


            iter = newIter;
            if( newIter != path.end() )
                ++ iter;
        }
        return 0;
    }

    ///====================================================================================================
    ///@brief Applies a function to all file with a given extension in a given directory.
    ///       Usage: 
    ///         for_all_files_with_extension("/home/gajdi/Desktop/Data/bionano/260", ".cmap", 
    ///            [](const auto& p)
    ///            { 
    ///                std::cout << "Found text file at " << p << "\n"; 
    ///            }
    ///         );
    ///
    ///@tparam FT   Function type.
    ///@param dir   Target directory
    ///@param ext   File extension
    ///@param f     pointer to a function
    ///====================================================================================================
    template <typename FT>
    void forFilesWithExtension(std::string_view dir, std::string_view ext, FT&& f)
    {
        for(const auto& p : fs::directory_iterator(dir))
        {
            if(p.path().extension() == ext)
            { 
                f(p);
            }
        }
    }

    template <typename FT>
    void forAllFilesWithExtension(std::string_view dir, std::string_view ext, FT&& f)
    {
        for(const auto& p : fs::recursive_directory_iterator(dir))
        {
            if(p.path().extension() == ext)
            { 
                f(p);
            }
        }
    }


}