#pragma once

#include <cstdint>
#include <string>
#include <string_view>

#include <functional>

namespace pg::parsing
{
	/// <summary> Function template allias </summary>
	///<template name="T">Ouput data type</template>
	template<typename T>
	using parseFn = std::function< void(const char* start, char* end, T& value)>;

	////////////////////////////////////////////////////////////////////////////////////////////////////
	/// <summary>	Calling parse function. </summary>
	/// <remarks>	Gajdi, 11.12.2019. </remarks>
	///<template name="T">Ouput data type</template>
	/// <param name="start">Start pointer of the string value to be parsed. </param>
	/// <param name="end">End pointer of the string value to be parsed. </param>
	/// <returns>TRUE if parsed, otherwise FALSE</returns>
	////////////////////////////////////////////////////////////////////////////////////////////////////
	template<typename T> inline bool parse(const char* start, char* end, T& value, parseFn<T> parseFunction)
	{
		return parseFunction(start, end, value);
	}

	//This templates will be specialized below
	template<typename T> constexpr void parseValue(const char* start, char* end, T& value) {}
	template<typename T> constexpr void parseValue(const std::string_view& delimiters, const char* start, char* end, T& value) {}

	template<> inline void parseValue(const char* start, char* end, float& value)
	{
		char tmp = *end;
		*end = '\0';
		value = static_cast<float>(atof(start));
		*end = tmp;
	}

	template<> inline void parseValue(const char* start, char* end, double &value)
	{
		char tmp = *end;
		*end = '\0';
		value = atof(start);
		*end = tmp;
	}

	template<> inline void parseValue(const char* start, char* end, std::int32_t &value)
	{
		char tmp = *end;
		*end = '\0';
		value = atoi(start);
		*end = tmp;
	}

	template<> inline void parseValue(const char* start, char* end, std::uint32_t &value)
	{
		char tmp = *end;
		*end = '\0';
		value = static_cast<std::uint32_t>(atoi(start));
		*end = tmp;
	}


	template<> inline void parseValue(const char* start, char* end, std::uint64_t &value)
	{
		char tmp = *end;
		*end = '\0';
		value = static_cast<std::uint64_t>(atoll(start));
		*end = tmp;
	}

	template<> constexpr void parseValue(const char* start, char* end, std::int8_t &value)
	{
		value = *start;
	}

	template<> constexpr void parseValue(const char* start, char* end, std::uint8_t &value)
	{
		value = static_cast<std::uint8_t>(*start);
	}

	template<> inline void parseValue(const char* start, char* end, std::string &value)
	{
		value = std::string(start, static_cast<const char*>(end));
	}

	template<> inline void parseValue(const std::string_view& delimiters, const char* start, char* end, std::string& value)
	{
		auto second = std::find_first_of(const_cast<char*>(start), end, std::cbegin(delimiters), std::cend(delimiters));
		value = std::string(start, static_cast<const char*>(second));
	}

	inline uint32_t fast_atou(const char *str)
	{
		unsigned int val = 0;
		while(*str) {
			val = (val << 1) + (val << 3) + *(str++) - 48;
		}
		return val;
	}

}