#pragma once

namespace pg::slidingwindowutils
{   
    ///====================================================================================================
    ///@brief Returns the number of windows after splitting |length| by a given sliding window.
    ///       Let have a data of size of "length" that must be splitted into windows of size "slidingWindowSize".
    ///       The sliding window can be shifted by "slideSize".
    ///@tparam T 
    ///@param slidingWindowSize The size of sliding Window = the size of a window
    ///@param slideSize SlidingWindows shift size.
    ///@param length The total length to be splitted into windows.
    ///@return size_t The number of windows after splitting.
    ///====================================================================================================
    template<typename T>
    size_t getNumberOfWindows(const T slidingWindowSize, const T slideSize, const T length)
    {
        constexpr T zero = T(0);
        if ((length < zero) || (slidingWindowSize < zero) || (slideSize <= zero) || (slidingWindowSize < slideSize)) return 0;
        if (length < slidingWindowSize) return 1;

        T x = (length - slidingWindowSize)/slideSize + 1;
        
        //VERSION A: If value = n*windowsSize, this value is a left bound of the next sliding window and falls into this sliding window
        //			 E.g. value =100, windowSize = 25:  100/25 = 4 but 100 is a left bound od the 5-th sliding window... +1
        return static_cast<size_t>(x+1);
        
        //VERSION B:  If value = n*windowsSize, this value is a right bound of the current sliding window
        //			  E.g. value =100, windowSize = 25:  100/25 = 4 and 100 is a right bound od the 4-th sliding window
        // auto const t {std::trunc(x)};
        // return static_cast<size_t>(t + (t < x));
    }

    ///====================================================================================================
    ///@brief Returns the index of the first part (sliding window) to which the value falls in.
    ///       Let have a data of size of "length" that is splitted into parts of size "slidingWindowSize".
    ///       The sliding window was shifted by "slideSize". 
    ///       It is supposed the the most left value is 0.
    ///@tparam T 
    ///@param slidingWindowSize The size of sliding Window = the size of a part
    ///@param slideSize SlidingWindows shift size.
    ///@param value The value for which the sliding window index must be found.
    ///@return size_t Index of the first part (sliding window) to which the value falls in.
    ///====================================================================================================
	template<typename T>
    size_t getFirstWindowIdx(const T slidingWindowSize, const T slideSize, const T value)
    {
        constexpr T zero = T(0);
        if ((value < zero) || (value < slidingWindowSize) || (slidingWindowSize < zero) || (slideSize <= zero) || (slidingWindowSize < slideSize)) return 0;
        return static_cast<size_t>((value-slidingWindowSize)/slideSize) + 1;
    }

    ///====================================================================================================
    ///@brief Get the UPPER BOUND of the number of Sliding windows that intersect with any value with respect to
    ///       windowSize and slideSize
    ///       E.g.:     |--------------|        slidingWindowSize = 6, slideSize = 3  => 6/3 = 2
    ///                 |------|
    ///                     |------|
    ////                        |------|
    ////                            |------|
    ///@tparam T 
    ///@param slidingWindowSize The size of sliding Window = the size of a part
    ///@param slideSize SlidingWindows shift size.
    ///@return size_t 
    ///====================================================================================================
	template<typename T>
    size_t getMaxNumberOfIntersectingWindows(const T slidingWindowSize, const T slideSize)
    {
        return static_cast<size_t>(slidingWindowSize/slideSize)+1;
    }

	template<typename T>
    T getWindowStart(const T slideSize, const size_t windowIdx)
    {
        return static_cast<T>(windowIdx) * slideSize;
    }

    template<typename T>
    T getWindowEnd(const T slidingWindowSize, const T slideSize, const size_t windowIdx)
    {
        return static_cast<T>(windowIdx) * slideSize + slidingWindowSize;
    }


    ///====================================================================================================
    ///@brief Get the number of Sliding windows that intersect with any value with respect to
    ///       windowSize and slideSize. Is is supposed that slideSize < windowSize to prevent gaps.
    ///       E.g.:     |--------------|        slidingWindowSize = 6, slideSize = 3  => 6/3 = 2
    ///                 |------|
    ///                     |------|
    ////                        |------|
    ////                            |------|
    ///@tparam T 
    ///@param slidingWindowSize The size of sliding Window = the size of a part
    ///@param slideSize SlidingWindows shift size. Must be < slidingWindowSize to prevent gaps.
    ///@param length The total length
    ///@param value The value for which the computation will be processed
    ///@return size_t 
    ///====================================================================================================
	template<typename T>
    size_t getNumberOfIntersectingWindows(const T slidingWindowSize, const T slideSize, const T length, const T value)
    {
        size_t firstIdx = getFirstWindowIdx(slidingWindowSize, slideSize, value);
        T startPos = static_cast<T>(firstIdx) * slideSize;
        T endPos = startPos + slidingWindowSize;
        size_t rv = 0;
        while ((startPos <= value)&&(value < endPos)&&(endPos < length))
        {
            ++rv;
            startPos = static_cast<T>(++firstIdx) * slideSize;
            endPos = startPos + slidingWindowSize;
        }
        return rv;
    }
}