#pragma once

#include <type_traits>
#include <iostream>
#include <fstream>
#include <tuple>
#include <utility>
#include <string_view>
#include <stdexcept>
#include <memory>

#include "parsing.h"
#include "serializing.h"

#include <utils/stringUtils.h>

#define NOINLINE __attribute__((noinline))
#define ALWAYS_INLINE __attribute__((always_inline))

namespace pg::stdtupleutils
{   
    ///====================================================================================================
    ///@brief Check if type is tuple
    ///====================================================================================================
    template <typename> struct is_tuple: std::false_type {};
    template <typename ...T> struct is_tuple<std::tuple<T...>>: std::true_type {};


    ///====================================================================================================
    ///@brief Applies a function to all elements of a tuple
    ///       E.g.: 
    ///             auto printer = [](auto idx, auto el) 
    ///             {
    ///                os << (idx == 0 ? "" : ", ") << el;
    ///                return 0;
    ///             };
    ///             tuple_for_each(tuple, printer);
    ///
    ///@tparam Func 
    ///@tparam Tuple 
    ///@tparam I 
    ///@param tuple 
    ///@param f 
    ///====================================================================================================
    template <typename Func, typename Tuple, std::size_t... I>
    decltype(auto) apply_for_each_impl(Tuple&& tuple, Func&& f, std::index_sequence<I...>)
    {
        auto swallow = { (std::forward<Func>(f)(I, std::get<I>(std::forward<Tuple>(tuple))))... };
        (void)swallow;      //ensures that the variable will be used 
    }

    template <typename Func, typename... Args>
    decltype(auto) apply_for_each(std::tuple<Args...>& tuple, Func&& f)
    {
        apply_for_each_impl(tuple, f, std::index_sequence_for<Args...>{});
    }
    

    ///====================================================================================================
    ///@brief Helper functions to parse Tuples
    ///====================================================================================================
    template<class Tuple, unsigned S>
    struct parse_by_idx
    {
        static void parseTuple(char* lineStart, char* lineEnd, Tuple& t, unsigned idx) ALWAYS_INLINE 
        {
            if (S == 0) return;
            constexpr unsigned C = S-1;

            if (idx == C) 
            {
                //typedef typename std::tuple_element_t<C,Tuple> elem_type;
                pg::parsing::parseValue(lineStart, lineEnd, std::get<C>(t));
            }
            else 
            {
                parse_by_idx<Tuple,C>::parseTuple(lineStart, lineEnd, t, idx);
            }
        }
    };
    template<class Tuple>
    struct parse_by_idx<Tuple, 0>		//Needed because of recursion!!!
    {
        static void parseTuple(char* lineStart, char* lineEnd, Tuple& t, unsigned idx) NOINLINE
        {
            throw std::runtime_error("bad field index");
        }
    };

    template<class Tuple>
    void parseTuple(char* lineStart, char* lineEnd, Tuple& t, unsigned idx) 
    {
        constexpr unsigned S = std::tuple_size_v<Tuple>;
        parse_by_idx<Tuple,S>::parseTuple(lineStart, lineEnd, t, idx);
    }

    template<typename... Ts>
    inline bool parseTuple(const std::string_view& delims, char* lineStart, char* lineEnd, std::tuple<Ts...>& r)
    {
        try 
        {
            const unsigned int tupleSize = static_cast<unsigned int>(std::tuple_size_v<std::tuple<Ts...>>);
            unsigned int recordItemID = 0;
            for (auto first = lineStart, second = lineStart, last = lineEnd; second != last && first != last; first = second + 1) 
            {
                second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
                if (first != second)
                {
                    parseTuple(first, second, r, recordItemID);
                    ++recordItemID;
                    if (recordItemID == tupleSize) break;
                }
            }
            return true;
        }
        catch (const std::exception& e)
        {
            std::cerr << e.what() << std::endl;
        }
        return false;
    }


    template<typename... Ts>
    inline bool parseSubTuple(const std::string_view& delims, 
                              const std::vector<uint32_t>& ascendingIndices,
                              char* lineStart, 
                              char* lineEnd, 
                              std::tuple<Ts...>& r)
    {
        try 
        {
            const unsigned int tupleSize = static_cast<unsigned int>(std::tuple_size_v<std::tuple<Ts...>>);
            unsigned int recordItemID = 0;
            unsigned int indicesID = 0;

            for (auto first = lineStart, second = lineStart, last = lineEnd; second != last && first != last; first = second + 1) 
            {
                second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
                if (first != second)
                {
                    if (ascendingIndices[indicesID] == recordItemID)
                    {
                        parseTuple(first, second, r, indicesID);
                        ++indicesID;
                        if (indicesID == ascendingIndices.size()) return true;
                    }
                    ++recordItemID;
                    if (recordItemID == tupleSize) break;
                }
            }
            return true;
        }
        catch (const std::exception& e)
        {
            std::cerr << e.what() << std::endl;
        }
        return false;
    }

    ///====================================================================================================
    ///@brief Recursive print_tuple function. The recursion solves std::get problem, because std::get needs to know an index in compile time !!!
    ///====================================================================================================
    template<std::size_t> struct int_{};

    template <class Tuple, size_t Pos>
    std::ostream& print_tuple(std::ostream& s, const Tuple& t, int_<Pos> ) 
    {
        s << std::get<std::tuple_size_v<Tuple> - Pos>(t) << ',';
        return print_tuple(s, t, int_<Pos-1>());
    }

    template <class Tuple>
    std::ostream& print_tuple(std::ostream& s, const Tuple& t, int_<1> ) 
    {
        return s << std::get<std::tuple_size_v<Tuple> - 1>(t);
    }

    template <class... Args>
    std::ostream& operator<<(std::ostream& s, const std::tuple<Args...>& t) 
    {
        s << '('; 
        print_tuple(s, t, int_<sizeof...(Args)>()); 
        return s << ')';
    }

    ///====================================================================================================
    ///@brief Final code for a constant-time function to retrieve an item from a tuple based on a runtime index. !!!
    /// It uses a look-up table. Untortunately, void* is returned which implies one need to know the target data type :-(
    ///====================================================================================================
    template <size_t ... Indexes, class Tuple>
    void* get_element_pointer(std::index_sequence<Indexes...>, Tuple & t, size_t idx) 
    {
        static void* ptrs[] = { const_cast<void*>(static_cast<const void*>(std::addressof(std::get<Indexes>(t))))... };
        return ptrs[idx];
    }

    template<class Tuple> __attribute__((noinline))
    constexpr void* get_element_pointer(Tuple& t, size_t index)
    {
        return get_element_pointer(std::make_index_sequence<std::tuple_size<Tuple>::value>(), t, index);
    }

    ///====================================================================================================
    ///@brief Better solution than above, unfortunately for tuples of the same type only, e.g. tuple<int, int, int>.
    /// The lookup table is te same type as the first element in the tuple ... see "return_type=typename std::tuple_element_t<0,Tuple>&;"
    ///====================================================================================================
    template<typename Tuple, typename Indices=std::make_index_sequence<std::tuple_size_v<Tuple>>>
    struct runtime_get_func_table;

    template<typename Tuple,size_t ... Indices>
    struct runtime_get_func_table<Tuple,std::index_sequence<Indices...>>
    {
        using return_type=typename std::tuple_element_t<0,Tuple>&;
        using get_func_ptr=return_type (*)(Tuple&) noexcept;
        static constexpr get_func_ptr table[std::tuple_size_v<Tuple>]={ &std::get<Indices>... };
    };

    template<typename Tuple,size_t ... Indices>
    constexpr typename 
    runtime_get_func_table<Tuple,std::index_sequence<Indices...>>::get_func_ptr
    runtime_get_func_table<Tuple,std::index_sequence<Indices...>>::table[std::tuple_size_v<Tuple>];

    template<typename Tuple>
    constexpr
    typename std::tuple_element_t<0,typename std::remove_reference_t<Tuple>>&
    runtime_get(Tuple&& t,size_t index)
    {
        using tuple_type=typename std::remove_reference_t<Tuple>;
        if(index>=std::tuple_size_v<tuple_type>)
            throw std::runtime_error("Out of range");
        return runtime_get_func_table<tuple_type>::table[index](t);
    }

    template<std::size_t> struct tsize{};
        
    ///====================================================================================================
    ///@brief Recursive tuple serialization
    ///====================================================================================================
    template <class Tuple, size_t Pos>
    void serialize_tuple(std::ofstream& s, const Tuple& t, tsize<Pos> ) 
    {
        constexpr unsigned C = std::tuple_size_v<Tuple> - Pos; 
        typedef typename std::tuple_element_t<C,Tuple> elem_type;
        if (std::is_same_v<elem_type,std::string>)
        {
            pg::serializing::serialize(std::get<C>(t), s);
        }
        else
        {
            s.write(reinterpret_cast<const char*>(&(std::get<C>(t))), sizeof(elem_type));
        }
        
        serialize_tuple(s, t, tsize<Pos-1>());
    }

    template <class Tuple>
    void serialize_tuple(std::ofstream& s, const Tuple& t, tsize<1> ) 
    {
        constexpr unsigned C = std::tuple_size_v<Tuple> - 1; 
        typedef typename std::tuple_element_t<C,Tuple> elem_type;
        if (std::is_same_v<elem_type,std::string>)
        {
            pg::serializing::serialize(std::get<C>(t), s);
        }
        else
        {
            s.write(reinterpret_cast<const char*>(&(std::get<C>(t))), sizeof(elem_type));
        }
    }

    template<typename T> 
    std::enable_if_t<is_tuple<T>::value, void>
    serializeTuple(const T& value, std::ofstream& s) 
    {
        constexpr unsigned S = std::tuple_size_v<T>;
        serialize_tuple(s, value, tsize<S>());
    }

     ///====================================================================================================
    ///@brief Recursive tuple deserialization
    ///====================================================================================================
    template <class Tuple, size_t Pos>
    void deserialize_tuple(std::ifstream& s, Tuple& t, tsize<Pos> ) 
    {
        constexpr unsigned C = std::tuple_size_v<Tuple> - Pos; 
        typedef typename std::tuple_element_t<C,Tuple> elem_type;
        if (std::is_same_v<elem_type,std::string>)
        {
            pg::serializing::deserialize(std::get<C>(t), s);
        }
        else
        {
            s.read(reinterpret_cast<char*>(&(std::get<C>(t))), sizeof(elem_type));
        }
        
        deserialize_tuple(s, t, tsize<Pos-1>());
    }

    template <class Tuple>
    void deserialize_tuple(std::ifstream& s, Tuple& t, tsize<1> ) 
    {
        constexpr unsigned C = std::tuple_size_v<Tuple> - 1; 
        typedef typename std::tuple_element_t<C,Tuple> elem_type;
        if (std::is_same_v<elem_type,std::string>)
        {
            pg::serializing::deserialize(std::get<C>(t), s);
        }
        else
        {
            s.read(reinterpret_cast<char*>(&(std::get<C>(t))), sizeof(elem_type));
        }
    }

    template<typename T> 
    std::enable_if_t<is_tuple<T>::value, void>
    deserializeTuple(T& value, std::ifstream& s) 
    {
        constexpr unsigned S = std::tuple_size_v<T>;
        deserialize_tuple(s, value, tsize<S>());
    }

}