#pragma once

#include <iostream>
#include <fstream>
#include <istream>
#include <unordered_set>
#include <vector>
#include <string_view>
#include <algorithm>

#include "parsing.h"

namespace pg::stdvectorutils
{   
    ///====================================================================================================
    ///@brief Usage: auto filteredVec = filter(myVec, [](int a) { return a > 5; });    where std::vector<int> mVec = {1,4,7,8,9,0};
    ///====================================================================================================
    template <typename Cont, typename Pred>
    void filter(const Cont &src, Pred predicate, Cont &dst) 
    {
        std::copy_if(src.begin(), src.end(), std::back_inserter(dst), predicate);
    }

    template<typename T>
    bool copyValues(const std::vector<T>& src, uint32_t from, uint32_t count, std::vector<T>& dst)
    {
        try
        {
            typename std::vector<T>::const_iterator start = src.begin() + from;
            typename std::vector<T>::const_iterator end = start + count;
            if (end > src.cend()) return false;
            dst.reserve(count);
            dst.assign(start,end);              //assign copy of the interval ... [start, end)    The last is not present!!!
            return true;
        }
        catch(const std::exception& e)
        {
            std::cerr << e.what() << std::endl;
        }
        return false;
    }

    template<typename T>
    bool serialize(const char* fileName, const std::vector<T>& v)
    {
        try
        {
            std::ofstream f(fileName, std::ios::binary);
            if (!f.is_open())
            {
                return false;
            }

            auto size = v.size();
            f.write(reinterpret_cast<char const*>(&size), sizeof(size));
            f.write(reinterpret_cast<char const*>(v.data()), v.size() * sizeof(T));

            f.flush();
            f.close();
            return true;
        }
        catch(const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    template<typename T>
    bool deSerialize(const char* fileName, std::vector<T>& v)
    {
        try
        {
            std::ifstream f(fileName, std::ios::in | std::ios::binary);
            if (!f.is_open())
            {
                return false;
            }

            decltype(v.size()) size;

            f.read(reinterpret_cast<char*>(&size), sizeof(size));
            v.resize(size);
            f.read(reinterpret_cast<char*>(v.data()), size * sizeof(T));
    
            f.close();
            return true;
        }
        catch(const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    template<typename T>
    bool serialize(const char* fileName, const std::vector<std::vector<T>>& v)
    {
        try
        {
            std::ofstream f(fileName, std::ios::binary);
            if (!f.is_open())
            {
                return false;
            }

            // write size of outer vector
            auto size_v = v.size();
            f.write(reinterpret_cast<char const*>(&size_v), sizeof(size_v));

            // write inner vectors
            for (auto& v_it : v)
            {
                auto size = v_it.size();
                f.write(reinterpret_cast<char const*>(&size), sizeof(size));
                f.write(reinterpret_cast<char const*>(v_it.data()), v_it.size() * sizeof(T));
            }

            f.flush();
            f.close();
            return true;
        }
        catch(const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    template<typename T>
    bool deSerialize(const char* fileName, std::vector<std::vector<T>>& v)
    {
        try
        {
            std::ifstream f(fileName, std::ios::in | std::ios::binary);
            if (!f.is_open())
            {
                return false;
            }

            // read size of outer vector and resize it
            decltype(v.size()) size_v;
            f.read(reinterpret_cast<char*>(&size_v), sizeof(size_v));
            v.resize(size_v);            

            decltype(v.size()) size;
            for (size_t s = 0; s < size_v; s++)
            {
                f.read(reinterpret_cast<char*>(&size), sizeof(size));
                v[s].resize(size);
                f.read(reinterpret_cast<char*>(v[s].data()), size * sizeof(T));
            }
    
            f.close();
            return true;
        }
        catch(const std::exception& e)
        {
            std::cout << e.what() << std::endl;
        }
        return false;
    }

    ///====================================================================================================
    ///@brief Parse vector of T from string data. Values are 
    ///@tparam T 
    ///@param delims Delimiters
    ///@param lineStart Line start
    ///@param lineEnd   Line end
    ///@param r Result vector
    ///@return bool TRUE if OK, otherwise FALSE.
    ///====================================================================================================
    template<typename T>
    bool parseVector(const std::unordered_set<char>& delims, char* lineStart, char* lineEnd, std::vector<T>& r)
    {
        try 
        {
            std::unordered_set<char>::const_iterator separatorIter;
            T tmp;
            for (auto first = lineStart, last = first; last != lineEnd && first != lineEnd; first = last + 1) 
            {
                last = first;
                while(last < lineEnd)
                {
                    separatorIter = delims.find(*last);
                    if (separatorIter != delims.end()) break;		//if (*it == delimiter)
                    last++;
                }
                if (first != last)
                {
                    pg::parsing::parseValue(static_cast<const char*>(first), last, tmp);
                    r.emplace_back(tmp);
                }
            }
            r.shrink_to_fit();
            return true;
        }
        catch (const std::exception& e)
        {
            std::cerr << e.what() << std::endl;
        }
        return false;
    }

    ///====================================================================================================
    ///@brief Parse vector of T from string data. Values are 
    ///@tparam T 
    ///@param delims string_view
    ///@param lineStart Line start
    ///@param lineEnd   Line end
    ///@param r Result vector
    ///@return bool TRUE if OK, otherwise FALSE.
    ///====================================================================================================
    template<typename T>
    bool parseVector(const std::string_view& delims, char* lineStart, char* lineEnd, std::vector<T>& r)
    {
        try 
        {
            T tmp;
            for (auto first = lineStart, second = lineStart, last = lineEnd; second != last && first != last; first = second + 1) 
            {
                second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
                if (first != second)
                {
                    pg::parsing::parseValue(first, second, tmp);
                    r.emplace_back(tmp);
                }
            }
            r.shrink_to_fit();
            return true;
        }
        catch (const std::exception& e)
        {
            std::cerr << e.what() << std::endl;
        }
        return false;
    }

    ///====================================================================================================
    ///@brief Splits a string_view data according to delimiters and returns separated parts
    ///@param delims Delimiters
    ///@param lineStart Line start
    ///@param lineEnd   Line end
    ///@param r Result vector
    ///====================================================================================================
    inline void splitSVPtr(std::string_view str, std::string_view& delims, std::vector<std::string_view>& parts)
    {
        parts.clear();
        for (auto first = str.data(), second = str.data(), last = first + str.size(); second != last && first != last; first = second + 1) 
        {
            second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
            if (first != second)
                parts.push_back({first, static_cast<std::string_view::size_type>(second - first)});
        }
    }

    ///====================================================================================================
    ///@brief Splits and parse a string_view data according to delimiters and returns separated items of T
    ///@tparam T Expected data type
    ///@param delims Delimiters
    ///@param lineStart Line start
    ///@param lineEnd   Line end
    ///@param r Result vector
    ///====================================================================================================
    template<typename T>
    void parseAddSplitSVPtrData(const std::string_view str, const std::string_view& delims, const uint32_t maxCount, std::vector<T>& r)
    {
        T tmp;
        for (auto first = str.data(), second = str.data(), last = first + str.size(); second != last && first != last; first = second + 1) 
        {
            second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
            if (first != second)
            {
                pg::parsing::parseValue(first, second, tmp);
                r.emplace_back(tmp);
                if (r.size() == maxCount) break;
            }
        }
    }

}