#pragma once

#include <cstdint>
#include <string_view>
#include <algorithm>
#include <unordered_set>

namespace pg::stringutils
{   
    inline bool getPart(const char* lineStart, const char* lineEnd, const std::string_view& delims, const std::uint32_t idx, std::string_view& part)
    {
        std::uint32_t counter = 0;
        for (auto first = lineStart, second = lineStart, last = lineEnd; second != last && first != last; first = second + 1) 
        {
            second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
            if (first != second)
            {
                 if (counter == idx)
                {
                    part = {first, static_cast<std::string_view::size_type>(second - first)};
                    return true;
                }
                ++counter;
            }
        }
        return false;
    }

    ///====================================================================================================
    ///@brief Splits a string_view data according to delimiters and returns idx-th part if exists. 
    ///       This method is almost 8x faster then with the use of unordered_set in case of a single delimiter.
    ///       In case of four delimiters, this method is 3-5x faster then with the use of unordered_set.
    ///@param str Text to be splitted
    ///@param delims StringView containing all possible delimiters
    ///@param idx Index of the requested part
    ///@param part String_view representing the searched part
    ///@return bool Returns TRUE if idx-th part exists, otherwise returns FALSE.
    ///====================================================================================================
    inline bool getPart(const std::string_view str, const std::string_view& delims, const std::uint32_t idx, std::string_view& part)
    {
        std::uint32_t counter = 0;
        for (auto first = str.data(), second = str.data(), last = first + str.size(); second != last && first != last; first = second + 1)
        {
            second = std::find_first_of(first, last, std::cbegin(delims), std::cend(delims));
            if (first != second)
            {
                if (counter == idx)
                {
                    part = {first, static_cast<std::string_view::size_type>(second - first)};
                    return true;
                }
                ++counter;
            }
        }
        return false;
    }

    ///====================================================================================================
    ///@brief Splits a string_view data according to delimiters and returns idx-th part if exists.
    ///@param str Text to be splitted
    ///@param delims Unordered set containing all possible delimiters
    ///@param idx Index of the requested part
    ///@param part String_view representing the searched part
    ///@return bool Returns TRUE if idx-th part exists, otherwise returns FALSE.
    ///====================================================================================================
    inline bool getPart(const std::string_view str, const std::unordered_set<char>& delims, const std::uint32_t idx, std::string_view& part)
    {
        std::uint32_t counter = 0;
        auto start = str.end();
        std::unordered_set<char>::const_iterator iter;

        for (auto it = str.begin(); it != str.end(); ++it)
        {
            iter = delims.find(*it);
            if (iter == delims.end())   //if (*it != delimiter)
            {
                if (start == str.end())
                    start = it;
                continue;
            }
            if (start != str.end()) 
            {
                if (counter == idx)
                {
                    part = {start, static_cast<std::string_view::size_type>(it - start)};
                    return true;
                }
                ++counter;
                start = str.end();
            }
        }
        if (start != str.end())
        {
            if (counter == idx)
            {
                part = {start, static_cast<std::string_view::size_type>(str.end() - start)};
                return true;
            }
        }
        return false;
    }

}